/* ### CHECK-ME-BEFORE-RELEASE ### */
/*
 * Title:   win32sound.cpp
 * Created: Spring 2000
 * Authors: Markus Noga
 *     $Id: win32sound.cpp,v 1.3 2003/11/19 10:37:31 weyrich Exp $
 *
 * Copyright (c) 2000--2003
 * Gilbert Baumann <unk6@stud.uni-karlsruhe.de>
 * Markus Noga <markus@noga.de>
 * Tim Weyrich <weyrich@inf.ethz.ch>
 *
 * This file is part of the Pointshop3D system.
 * See http: *www.pointshop3d.com/ for more information.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 * 
 * Contact info@pointshop3d.com if any conditions of this
 * licensing are not clear to you.
 *
 * ---------------------------------------------------------------
 *
 * $Log: win32sound.cpp,v $
 * Revision 1.3  2003/11/19 10:37:31  weyrich
 * added Pointshop3D headers
 *
 *
 */

#include <pisound.h>
#include <mmreg.h>
#include <mmsystem.h>

#define MINIMUM_TIMER_INTERVAL	20    // in ms.

//#define SHOW_LAG		        // show current output lag?

static void CALLBACK win32sound_timerCallback(UINT wTimerID, UINT msg, 
				      DWORD dwUser, DWORD dw1, DWORD dw2) {
  Win32Sound *snd=(Win32Sound*) dwUser;
  snd->timerCallback();
}

  
Win32Sound::Win32Sound(const char *dev,
        	       uint32 ra, uint32 bi, uint32 ch,
        	       uint32 bufsz, uint32 buf)
  : IfSound(dev,ra,bi,ch,bufsz,buf)
{
  DSBUFFERDESC bufDesc;
  DSBCAPS      bufCaps;
  WAVEFORMATEX fmt;
  HRESULT res;
  
  // check superclass initialization validity
  //
  if(!isValid())
    return;
  valid=false;

  // determine intended timer interval from buffer sizes.
  //
  timerId=NULL;
  timerInt=(1000*bufSize*(buffers-1))/rate;
  if(timerInt<MINIMUM_TIMER_INTERVAL)
    timerInt=MINIMUM_TIMER_INTERVAL;

#ifdef SHOW_LAG
  fprintf(stderr,"using %ld ms (^ %ld samples) timer interval.\n",
    timerInt,(timerInt*rate)/1000);
#endif

  // create DirectSound device
  //
  res=DirectSoundCreate(NULL,&ds,NULL);
  if(res!=DS_OK || ds==NULL) {
    IfSound::genError("creating");
    return;
  }
  ds->Initialize(NULL);

  // FIXME: find better way to obtain window handle!
  //
  HWND hWnd=GetForegroundWindow();
  if(hWnd==NULL) {
    hWnd=GetDesktopWindow();
  }
 
  // Set priority level (primary buffer if possible)
  //
  res=ds->SetCooperativeLevel(hWnd,DSSCL_WRITEPRIMARY);
  primary= res==DS_OK;
  if(!primary) {
    res=ds->SetCooperativeLevel(hWnd,DSSCL_EXCLUSIVE);
    if(res!=DS_OK) {
      IfSound::genError("SetCooperativeLevel");
      ds->Release();
      return;
    }
  }
#ifdef SHOW_LAG
  fprintf(stderr,"%susing primary buffer.\n", primary ? "" : "not ");
#endif

  // prepare buffer format
  //
  fmt.wFormatTag     =WAVE_FORMAT_PCM;
  fmt.nSamplesPerSec =rate;
  fmt.wBitsPerSample =(WORD) bits;
  fmt.nChannels      =(WORD) channels;
  fmt.nBlockAlign    =(WORD) getSampleSize();
  fmt.nAvgBytesPerSec=rate*getSampleSize();
  fmt.cbSize         =0;

  // allocate sound buffer
  //
  bufDesc.dwSize       =sizeof(DSBUFFERDESC);

  bufDesc.dwFlags      =(primary ? DSBCAPS_PRIMARYBUFFER : 0 ) |
			DSBCAPS_STICKYFOCUS   |
			DSBCAPS_GETCURRENTPOSITION2;
  bufDesc.dwBufferBytes=(primary ? 0 : sampleSize*((4*timerInt*rate)/1000) );
  bufDesc.dwReserved   =0;
  bufDesc.lpwfxFormat  =(primary ? NULL : &fmt);

  res=ds->CreateSoundBuffer(&bufDesc,&buffer,NULL);
  if(res!=DS_OK || buffer==NULL) {
    IfSound::genError("CreateSoundBuffer");
    ds->Release();
    return;
  }

  // set buffer format (for primary)
  //
  if(primary) {
    res=buffer->SetFormat(&fmt);
    if(res!=DS_OK) {
      IfSound::genError("SetFormat");
      buffer->Release();
      ds    ->Release();
      return;
    }
  }

  // determine buffer size
  //
  bufCaps.dwSize=sizeof(DSBCAPS);
  res=buffer->GetCaps(&bufCaps);
  if(res!=DS_OK || buffer==NULL) {
    IfSound::genError("GetBufCaps");
    buffer->Release();
    ds    ->Release();
    return;
  }
  bufSize=bufCaps.dwBufferBytes/getSampleSize();
  buffers=1;

  valid=true;
}

Win32Sound::~Win32Sound() {
  if(isValid()) {
    stop();
    buffer->Release();
    ds    ->Release();
  }
}

void Win32Sound::play() {
  if(isValid()) {
    HRESULT res;

    // pre-fill buffer at current position
    //
    bufPos=getWritePos();
    timerCallback();

    // play and start timer
    //
    res=buffer->Play(0,0,DSBPLAY_LOOPING);
    if(res==DS_OK)
      timerId=timeSetEvent(timerInt,timerInt/2,
                           win32sound_timerCallback,(DWORD) this,
                           TIME_PERIODIC | TIME_CALLBACK_FUNCTION);

    if(!timerId)
      buffer->Stop();
  }
}

void Win32Sound::stop() {
  if(isValid() && timerId) {
    timeKillEvent(timerId);
    timerId=NULL;
    buffer->Stop();
  }
}

void Win32Sound::timerCallback() {
  uint32  blockSize=(2*timerInt*rate)/1000;
  uint32  writePos;

  void   *start1, *start2;
  uint32  bytes1 , bytes2;
  HRESULT res;

  // compute remaining buffer to write
  //
  writePos=getWritePos();
  if(writePos<=bufPos)
    blockSize-=bufPos-writePos;
  else
    blockSize-=(bufPos+bufSize-writePos);

restored:
  res=buffer->Lock(bufPos*getSampleSize(),
                   blockSize*getSampleSize(),
                   &start1,&bytes1,
		   &start2,&bytes2,
		   0);

  // restore buffer when input focus was lost.
  //
  if(res==DSERR_BUFFERLOST) {
#ifdef SHOW_LAG
    fprintf(stderr,"lost ");
#endif
    res=buffer->Restore();
    if(res==DS_OK) {
#ifdef SHOW_LAG
      fprintf(stderr,"RESTORED ");
#endif
      if(res==DS_OK) {
        res=buffer->Play(0,0,DSBPLAY_LOOPING);
	if(res==DS_OK)
	  goto restored;
      }
    }
  }

  if(res==DS_OK) {
    uint32 size1=bytes1/getSampleSize();
    uint32 size2=bytes2/getSampleSize();

    renderBuffer(start1,size1);
    if(start2)
      renderBuffer(start2,size2);

#ifdef SHOW_LAG
    uint32 lag=(bufPos+bufSize-getPlayPos()) % bufSize;
    fprintf(stderr,"%ld ",lag);
#endif

    bufPos= (bufPos+size1+size2) % bufSize;
    res=buffer->Unlock(start1, bytes1, start2, bytes2);
  }
}

void Win32Sound::genError(const char *msg,MMRESULT nr) {
  if(nr) {
    char err[MAXERRORLENGTH];
    uint32 len;

    waveOutGetErrorText(nr,err,MAXERRORLENGTH);
    len=strlen(msg)+strlen(err)+10;

    errMsg=new char[len];
    if(errMsg)
      sprintf(errMsg,"sound: %s: %s",msg,err);
  } else
    IfSound::genError(msg);
}
