/* ### CHECK-ME-BEFORE-RELEASE ### */
/*
 * Title:   gff.c
 * Created: Fri Apr 12 11:19:09 2002
 * Authors: Tim Weyrich <weyrich@inf.ethz.ch>
 *
 * Copyright (c) 2001--2003, Computer Graphics Lab, ETH Zurich
 *
 * This file is part of the Pointshop3D system.
 * See http: *www.pointshop3d.com/ for more information.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA
 *
 * Contact info@pointshop3d.com if any conditions of this
 * licensing are not clear to you.
 *
 * ---------------------------------------------------------------
 *
 * $Log: gff.c,v $
 * Revision 1.10  2003/11/19 13:04:01  weyrich
 * added Pointshop3D headers
 *
 * Revision 1.9  2003/11/17 15:05:52  weyrich
 * new API allowing for more than 32 surfel properties
 *
 * Revision 1.8  2003/03/20 14:52:51  weyrich
 * smaller fixes of implicit coordinate system transformations
 *
 * Revision 1.7  2002/12/04 19:50:25  weyrich
 * removed C++ comments
 *
 * Revision 1.6  2002/12/03 14:20:22  weyrich
 * libpi needed libpthread
 *
 * Revision 1.5  2002/11/18 13:09:12  weyrich
 * added markers and tangential vectors
 *
 * Revision 1.4  2002/05/23 13:59:56  weyrich
 * First official release.
 *
 * Revision 1.3  2002/05/17 13:58:16  weyrich
 * PointShop3D integration nearly finished
 *
 * Revision 1.2  2002/05/08 09:20:35  weyrich
 * added PI_PACKED/pragma pack to disc structures
 *
 * Revision 1.1.1.1  2002/05/08 08:58:47  weyrich
 * surfel file format
 */

#include <pisystem.h>
#include <pistring.h>
#include <piendian.h>

#include <stddef.h>             /* for offsetof() */

#include <errno.h>
#include <assert.h>

#if 0
#  include "sfldefs.h"
#else
#  include "gff.h"
#endif

#define  D(d)   ((gff_dir*)(d))
#define  cD(d)  ((const gff_dir*)(d))

#define  TAG_IS_INLINED(t)  (((t)->type == GFFTYPE_2OFFSET                                      \
                              || (t)->length*gff_type_size[(t)->type] <= sizeof((t)->offset))   \
                             && (t)->type != GFFTYPE_LDB)
#define  TAG_LENGTH_VALID(t)  ((t)->type != GFFTYPE_2OFFSET && (t)->length != GFF_UNKNOWN_SIZE)

#define  NEVER_REACHED()  do{fprintf(stderr, "%s:%d: NEVER REACHED\n", __FILE__, __LINE__);}while(0)
#define  NEVER_REACHED_ERRNO()  do{fprintf(stderr, "%s:%d: NEVER REACHED: %s\n", __FILE__, __LINE__, strerror(errno));}while(0)

#define  HEAP_OVERFLOW()  do{fprintf(stderr, "%s:%d: HEAP OVERFLOW\n", __FILE__, __LINE__);}while(0)

#define  WRITE_ERROR()  do{fprintf(stderr, "%s:%d: write error: %s\n", __FILE__, __LINE__, strerror(errno));}while(0)
#define  READ_ERROR()  do{fprintf(stderr, "%s:%d: read error: %s\n", __FILE__, __LINE__, strerror(errno));}while(0)
#define  SEEK_ERROR()  do{fprintf(stderr, "%s:%d: couldn't seek: %s\n", __FILE__, __LINE__, strerror(errno));}while(0)

#define  GFF_FILE_PADDING       4
#define  GFF_FILE_PADDING_FLOOR(o) ((o) & ((uint32)(-(GFF_FILE_PADDING))))
#define  GFF_FILE_PADDING_CEIL(o)  (((o)+(GFF_FILE_PADDING-1)) & ((uint32)(-(GFF_FILE_PADDING))))
#define  GFF_FILE_PADDING_FRAC(o)  ((o) & (~(uint32)(-(GFF_FILE_PADDING))))

/* It is important, that the memory layout of gff_tag starts exactly
   the same as gff_tag_on_disk!
*/
typedef struct gff_tag
{
  uint16    id;
  uint16    type;
  uint32    length;
  uint32    offset;

  void     *values;

  struct gff_tag  *cdr;
  uint32           target_offset;
}
gff_tag;

#ifdef _WIN32
#  pragma pack(push, __pre_gff_tag_on_disk__)
#  pragma pack(4)
#endif

typedef struct
{
  uint16         id;
  uint16         type;
  uint32         length;
  uint32         offset;
}
gff_tag_on_disk PI_PACKED;

#ifdef _WIN32
#  pragma pack(pop, __pre_gff_tag_on_disk__)
#endif

#define  GFF_TAG_ON_DISK_SIZE  12

/*! \brief Internal directory representation.
 *
 *  Used to read/compose directories.
 * */
typedef struct gff_dir
{
  uint32    offset;
  int       num_tags;           /*!< \brief Number of tags. */
  int       num_children;       /*!< \brief Number of child nodes. */

  /* There is no 'parent_offset'. Instead next_offset_offset and
     num_children_offset are used. */
  struct gff_dir  *parent_dir;          /*!< \brief Parent directory or \e NULL. */

  struct gff_tag  *tags;                /*!< \brief Linked list of tags. */

  uint32          *child_offsets;
  struct gff_dir **child_dirs;
  uint32           next_offset;
  struct gff_dir  *next_dir;
  uint32           up_link_offset;
  struct gff_dir  *up_link_dir;

  uint32  next_offset_offset;   /*!< \brief File offset, where next_offset can be found. */
  uint32  num_children_offset;  /*!< \brief File offset, where num_children can be found. */

  int       cur_num_children;           /*!< \brief Number of child nodes that have already been processed. */
}
gff_dir;

/*****************************************************************************/

const char  *gff_tag_str[GFF_INTERNAL_MAXTAG] =
  {
    "UPLINK",
    "LINK",
    "DOWNLINK",
    "NUM_CHILDREN"
  };

const char  *gff_type_str[GFFTYPE_NUM_TYPES] =
  {
    "UINT8",
    "INT8",
    "UINT16",
    "INT16",
    "UINT32",
    "INT32",
    "FLOAT",
    "DOUBLE",
    "ASCII",
    "APPLICATION",
    "LDB",
    "OFFSET",
    "2OFFSET"
  };

/*****************************************************************************/

static gff_tag  *new_tag(void)
{
  gff_tag  *result;
  if ((result = (gff_tag *)calloc(1, sizeof(gff_tag))))
    {
      return result;
    }
  return NULL;
}

static void  free_tag(gff_tag *tag)
{
  if (tag)
    {
      if (tag->values && tag->values != &tag->length && tag->values != &tag->offset)
        free(tag->values);
      free_tag(tag->cdr);
      free(tag);
    }
}

static gff_tag  *find_tag(const gff_dir *dir, int tag_id)
{
  if (dir)
    {
      gff_tag  *result;
      for (result=dir->tags; result && result->id != tag_id; result=result->cdr);
      return result;
    }
  else
    {
      fprintf(stderr, "%s:%d:find_tag: Request for tag in directory, which is NULL.\n",
              __FILE__, __LINE__);
      return NULL;
    }
}

/*! \brief Appends a new tag to a directory.
 *
 *  If a tag with the same id already exists, 1 is returned.
 *
 *  The new tag's value pointer is set to a new allocated memory, or,
 *  if the data is small enough to be stored in-line, set to a
 *  position inside the tag. Except for type GFFTYPE_2OFFSET, no
 *  memory is allocated, if \a num_elements is 0.
 *
 *  On heap overflow, \a *t is set to NULL, and still 0 is returned.
 *
 *   \return            1 on tag collision; 0 else
 */
static int  create_tag(gff_dir *dir, int tag_id, int field_type, int num_elements, gff_tag **newtag)
{
  gff_tag  **slot;
  for (slot=&dir->tags; (*slot) && (*slot)->id != tag_id; slot=&(*slot)->cdr);
  if (*slot)
    return 1;
  else
    {
      if ((*newtag = ((*slot) = new_tag())))
        {
          (*newtag)->id = tag_id;
          (*newtag)->type = field_type;
          (*newtag)->length = num_elements;
          dir->num_tags++;
          if ((*newtag)->type == GFFTYPE_2OFFSET)
            {
              /* let data pointer point to (*newtag)->length */
              (*newtag)->values = &(*newtag)->length;
              return 0;
            }
          else if (num_elements > 0)
            {
              if ((*newtag)->length * gff_type_size[(*newtag)->type] <= sizeof((*newtag)->offset))
                {
                  /* let data pointer point to (*newtag)->offset */
                  (*newtag)->values = &(*newtag)->offset;
                  return 0;
                }
              else
                {
                  /* allocate memory for num_elements values */
                  (*newtag)->values = malloc((*newtag)->length * gff_type_size[(*newtag)->type]);
                  /* ### FIXME: tidy up! */
                  return !(*newtag)->values;
                }
            }
          else
            return 0;
        }
    }
  return 1;
}

static const char *gff_get_tag_string(uint32 id)
{
  if (id < GFF_INTERNAL_MAXTAG)
    return gff_tag_str[id];
  else
    return NULL;
}

static int  dump_tag(FILE *stream, const gff_tag *tag, const char *(*get_tag_string)(uint32 id))
{
  if (!get_tag_string)
    get_tag_string = gff_get_tag_string;
  
  if (tag->target_offset)
    fprintf(stream, "@0x%08x:  ", (unsigned)tag->target_offset);
  else
    fprintf(stream, "  ");

  if (get_tag_string(tag->id))
    fprintf(stream, "tag:%14s  ", get_tag_string(tag->id));
  else
    fprintf(stream, "tag:    0x%08x  ",
            (unsigned)tag->id);

  if (tag->type < GFFTYPE_NUM_TYPES)
    if (tag->type == GFFTYPE_2OFFSET)
      fprintf(stream, "type:%12s  length:    -  value[s]:",
              gff_type_str[tag->type]);
    else
      fprintf(stream, "type:%12s  length:%5d  value[s]:",
              gff_type_str[tag->type], (int)tag->length);
  else
    fprintf(stream, "type:  0x%08x  length:%5d  value[s]:",
            (unsigned)tag->type, (int)tag->length);

  if (tag->type == GFFTYPE_ASCII)
    fprintf(stream, " \"%s\"", (const char *)tag->values);
  else if (tag->type == GFFTYPE_2OFFSET)
    {   
      fprintf(stream, " 0x%08x 0x%08x",
              (unsigned)((uint32 *)tag->values)[0], (unsigned)((uint32 *)tag->values)[1]);
    }    
  else if (tag->length > 0)
    {
      int  i, n = tag->length < 5 ? tag->length : 5;

      for (i=0; i<n; i++)
        {
          if (i == 5-1)
            fprintf(stream, " ...");
          else
            switch(tag->type)
              {
              case GFFTYPE_UINT8:
                fprintf(stream, " %d", (int)((uint8 *)tag->values)[i]);
                break;
                
              case GFFTYPE_INT8:
                fprintf(stream, " %d", (int)((int8 *)tag->values)[i]);
                break;
                
              case GFFTYPE_UINT16:
                fprintf(stream, " %d", (int)((uint16 *)tag->values)[i]);
                break;
                
              case GFFTYPE_INT16:
                fprintf(stream, " %d", (int)((int16 *)tag->values)[i]);
                break;
                
              case GFFTYPE_UINT32:
              case GFFTYPE_OFFSET:
                fprintf(stream, " 0x%08x", (unsigned)((uint32 *)tag->values)[i]);
                break;
                
              case GFFTYPE_FLOAT:
                fprintf(stream, " %g", (double)((float *)tag->values)[i]);
                break;
                
              case GFFTYPE_DOUBLE:
                fprintf(stream, " %g", ((double *)tag->values)[i]);
                break;
              }
        }
    }
  fprintf(stream, "\n");
  return 0;
}

static gff_dir  *new_dir(void)
{
  gff_dir  *result;
  if ((result = (gff_dir *)calloc(1, sizeof(gff_dir))))
    {
      return result;
    }
  return NULL;
}

static void  free_dir(gff_dir *dir)
{
  if (dir)
    {
      int  i;
      
      for (i=0; i<dir->num_children; i++)
        if (dir->child_dirs[i])
          free_dir(dir->child_dirs[i]);
      
      if (dir->child_dirs)     free(dir->child_dirs);
      if (dir->child_offsets)  free(dir->child_offsets);

      free_tag(dir->tags);
      
      free(dir);
    }
}

static gff_file  *new_file(void)
{
  gff_file  *result;
  if ((result = (gff_file *)calloc(1, sizeof(gff_file))))
    {
      return result;
    }
  return NULL;
}

static void  free_file(gff_file *file)
{
  if (file)
    {
      free_dir(file->root_dir);
      free(file);
    }
}

/*****************************************************************************/

int32  gff_read(gff_file *file, void *buf, uint32 size)
{
  uint32  left = size;

  while(left)
    {
      int32  res = read (file->fd, buf, left);

      if (res < 0)
        return -1;
      else if (res == 0) /* End of file! */
        break;

      left -= res;
    }

  return size - left;
}

/* ------------------------------------------------------------------------- */

int32  gff_write(gff_file *file, const void *buf, uint32 size)
{
  uint32  left = size;

  while(left)
    {
      int32  res = write (file->fd, buf, left);

      if (res < 0)
        return -1;

      left -= res;
      buf = ((char *)buf) + res;
    }

  return size;
}

/*****************************************************************************/

static int  cmp_magic(uint32 magic, const char *magic_ref)
{
  return magic - *((const uint32*)magic_ref);
}

/*!   \param mode       "r" or "w"
 *    \return           gff file pointer; NULL on error or if file
 *                      for reading is not GFF
 **/
gff_file  *gff_open(const char *fname, const char *mode, const char magic[4], uint32 version)
{
  gff_file  *file = NULL;
  (void)dump_tag;
  if (strchr(mode, 'r'))
    { /* read */
      int  fd = open(fname, O_RDONLY|O_BINARY);
      if (fd >= 0)
        return gff_fdopen(fd, mode, magic, version);
      else
        fprintf(stderr, "gff_open: couldn't open `%s' for reading: %s\n",
                fname, strerror(errno));
    }
  else if (strchr(mode, 'w'))
    { /* write */
      int fd  = open(fname, O_CREAT|O_WRONLY|O_TRUNC|O_BINARY, 0666);
      if (fd >= 0)
        return gff_fdopen(fd, mode, magic, version);
      else
        fprintf(stderr, "gff_open: couldn't open `%s' for writing: %s\n",
                fname, strerror(errno));
    }
  else
    fprintf(stderr, "%s:%d:gff_open: invalid mode string `%s'\n", __FILE__, __LINE__, mode);

  if (file)
    free(file);
  
  return NULL;
}

/*! The descriptor will be closed, when gff_close() is called.
 *
 *    \param mode       "r" or "w"
 *    \return           gff file pointer; NULL on error or if file
 *                      for reading is not GFF
 **/
gff_file  *gff_fdopen(int fd, const char *mode, const char magic[4], uint32 version)
{
  if (sizeof(gff_header) != GFF_HEADER_SIZE)
    {
      fprintf(stderr, "%s:%d: FATAL: padding check failed for gff_header\n",
              __FILE__, __LINE__);
      return NULL;
    }
  else if (sizeof(gff_tag_on_disk) != GFF_TAG_ON_DISK_SIZE)
    {
      fprintf(stderr, "%s:%d: FATAL: padding check failed for gff_tag\n",
              __FILE__, __LINE__);
      return NULL;
    }
  else
    {
      gff_file  *file = NULL;
      if (strchr(mode, 'r'))
        { /* read */
          if ((file = new_file()))
            {
              file->fd = fd;
              
              file->root_dir = NULL;
              file->active_dir = NULL;
              file->read_p = 1;
                  
              gff_read(file, &file->hd, sizeof(file->hd));
              pi_uint32_from_net_ip(&file->hd.gff_version);
              pi_uint32_from_net_ip(&file->hd.inst_version);
              pi_uint32_from_net_ip(&file->hd.root);
              if (!cmp_magic(file->hd.gff_magic, GFF_MAGIC))
                {
                  if ((file->hd.gff_version & 0xff00) <= (GFF_VERSION & 0xff00))
                    {
                      if (!magic || !cmp_magic(file->hd.inst_magic, magic))
                        {
                          if ((file->hd.inst_version & 0xff00) <= (version & 0xff00))
                            {
                              if (lseek(file->fd, file->hd.root, SEEK_SET) >= 0)
                                {
                                  int  cont = 1;
                                      
                                  if (lseek(file->fd, 0, SEEK_SET) >= 0)
                                    {
                                      file->seekable_p = 1;
                                      if (lseek(file->fd, file->hd.root, SEEK_SET) < 0)
                                        {
                                          NEVER_REACHED_ERRNO();
                                          cont = 0;
                                        }
                                    }
                                  else
                                    file->seekable_p = 0;

                                  if (cont)
                                    {
                                      file->next_dir_to_read_offs = file->hd.root;
                                      file->next_dir_to_read_parent = NULL;
                                      return file;
                                    }
                                }
                              else
                                fprintf(stderr, "gff_open: couldn't seek to root directory: %s\n", strerror(errno));
                            }
                          else
                            fprintf(stderr, "gff_open: sorry, unknown major version of format instance...\n");
                        }
                      else
                        fprintf(stderr, "gff_open: sorry, no appropriate format instance...\n");
                    }
                  else
                    fprintf(stderr, "gff_open: sorry, unknown major GFF version...\n");
                }
              else
                fprintf(stderr, "gff_open: sorry, no GFF file...\n");
            }
          else
            fprintf(stderr, "%s:%d: HEAP OVERFLOW!\n", __FILE__, __LINE__);
        }
      else if (strchr(mode, 'w'))
        { /* write */
          if ((file = new_file()))
            {
              file->fd = fd;
                  
              file->hd.gff_magic = *((const uint32 *)GFF_MAGIC);
              file->hd.gff_version = GFF_VERSION;
              file->hd.inst_magic = magic ? *((const uint32 *)magic) : 0x0;
              file->hd.inst_version = version;
              file->hd.root = sizeof(file->hd);

              pi_uint32_to_net_ip(&file->hd.gff_version);
              pi_uint32_to_net_ip(&file->hd.inst_version);
              pi_uint32_to_net_ip(&file->hd.root);
              gff_write(file, &file->hd, sizeof(file->hd));
              pi_uint32_from_net_ip(&file->hd.gff_version);
              pi_uint32_from_net_ip(&file->hd.inst_version);
              pi_uint32_from_net_ip(&file->hd.root);

              file->root_dir = NULL;
              file->active_dir = new_dir();
              file->read_p = 0;

              {
                int  cont = 1;
                if (lseek(file->fd, 0, SEEK_SET) >= 0)
                  {
                    file->seekable_p = 1;
                    if (lseek(file->fd, file->hd.root, SEEK_SET) < 0)
                      {
                        SEEK_ERROR();
                        cont = 0;
                      }
                  }
                else
                  file->seekable_p = 0;

                if (cont)
                  return file;
              }  
            }
          else
            fprintf(stderr, "%s:%d: HEAP OVERFLOW!\n", __FILE__, __LINE__);
        }
      else
        fprintf(stderr, "%s:%d:gff_open: invalid mode string `%s'\n", __FILE__, __LINE__, mode);

      if (file)
        free(file);
  
      return NULL;
    }
}

static int  write_uplink(gff_file *file, int eof_tag_p);

int  gff_close(gff_file *file)
{
  int  result=0;
  if (!file->read_p)
    if (write_uplink(file, 1))
      {
        fprintf(stderr, "%s%d:gff_close: could not write eof tag\n", __FILE__, __LINE__);
        result = 1;
      }
  result |= close(file->fd);
  free_file(file);
  return result;
}


/*****************************************************************************
 *  Read mode
 *****************************************************************************/

uint32  gff_get_magic(const gff_file *file)
{
  return file->hd.inst_magic;
}

uint32  gff_get_version(const gff_file *file)
{
  return file->hd.inst_version;
}

int  gff_is_seekable(const gff_file *file)
{
  return file->seekable_p;
}

void  gff_assume_not_seekable(gff_file *file)
{
  file->seekable_p = 0;
}

/*****************************************************************************/

/*! Works for net->local and local->net, as well.
 */
static void  swap_data(int type, uint32 length, void *data)
{
  if (PI_LOCAL_ENDIANESS != PI_NETWORK_ENDIAN)
    {
      uint32  i;
      switch(gff_type_size[type])
        {
        case 1:  break;
        case 2:
          {
            uint16  *dat = (uint16 *)data;
            (void)dat;  /* to suppress non-usage warning on big-endian machines... */
            for (i=0; i<length; i++)
              pi_uint16_from_net_ip(&dat[i]);
          }
          break;
        case 4:
          {
            uint32  *dat = (uint32 *)data;
            (void)dat;  /* to suppress non-usage warning on big-endian machines... */
            for (i=0; i<length; i++)
              pi_uint32_from_net_ip(&dat[i]);
          }
          break;
        case 8:
          {
            double  *dat = (double *)data;
            (void)dat;  /* to suppress non-usage warning on big-endian machines... */
            for (i=0; i<length; i++)
              pi_double_from_net_ip(&dat[i]);
          }
          break;
        }
    }
}

/*! Doesn't read the data. It has to be read, if result->values is
 *  \e NULL.
 */
static gff_tag  *fetch_tag(gff_file *file)
{
  gff_tag  *result;
  if ((result = new_tag()))
    {
      int32          read_size;
      gff_tag_on_disk  tag;
      
      if ((read_size = gff_read(file, &tag, sizeof(tag))) > 0)
        {
          pi_uint16_from_net_ip(&tag.id);
          pi_uint16_from_net_ip(&tag.type);
          pi_uint32_from_net_ip(&tag.length);
          pi_uint32_from_net_ip(&tag.offset);

          result->id = tag.id;
          result->type = tag.type;
          result->length = tag.length;
          result->offset = tag.offset;

          if (result->type == GFFTYPE_2OFFSET)
            {
              result->values = &result->length;
            }
          else if (TAG_IS_INLINED(result))
            {
              /* twiddle offset back, since there might have been data, that
                 was swapped differently */
              pi_uint32_to_net_ip(&result->offset);

              /* let data pointer point to result->offset */
              result->values = &result->offset;

              /* swap data again */
              swap_data(result->type, result->length, result->values);
            }
          else
            result->values = NULL;
      
          return result;
        }
      else
        {
          if (read_size < 0)
            fprintf(stderr, "%s:%d:read: %s\n", __FILE__, __LINE__, strerror(errno));
          free (result);
          return NULL;
        }
    }
  return NULL;
}

static gff_dir  *fetch_dir(gff_file *file, gff_dir *parent_dir)
{
  gff_dir  *dir = new_dir();
#ifdef GFF_DEBUG_OUTPUT
  fprintf(stderr, "FETCH_DIR: 0x%08x\n", (unsigned)lseek(file->fd, 0, SEEK_CUR));
#endif
  if (dir)
    {
      uint32  data_offset = 0;
      /* ### Am I already at the right position? */
      dir->offset = lseek(file->fd, 0, SEEK_CUR);
      dir->parent_dir = parent_dir;
          
      /* Read tag definitions. */
      {
        gff_tag  **slot;
        int        reached_end_p = 0;

        for (slot=&dir->tags; !reached_end_p; slot=&(*slot)->cdr)
          {
            if (!(*slot = fetch_tag(file)))
              break;

            dir->num_tags++;

            switch ((*slot)->id)
              {
              case GFFTAG_UPLINK:
                assert((*slot)->values);
                assert(dir->tags == *slot);
                dir->up_link_offset = ((uint32*)(*slot)->values)[0];
                reached_end_p = 1;
                break;
                  
              case GFFTAG_LINK:
                assert((*slot)->values);
                dir->next_offset = ((uint32*)(*slot)->values)[0];
                
                if (dir->parent_dir &&
                    dir->parent_dir->num_children > dir->parent_dir->cur_num_children+1 &&
                    dir->next_offset)
                  dir->parent_dir->child_offsets[dir->parent_dir->cur_num_children+1] = dir->next_offset;
                
                reached_end_p = 1;
                
                break;
                  
              case GFFTAG_DOWNLINK:
                assert((*slot)->values);
                assert(((gff_tag *)(((char *)slot) - offsetof(gff_tag, cdr)))->id == GFFTAG_NUM_CHILDREN);
                    
                dir->next_offset = ((uint32*)(*slot)->values)[0];
                dir->child_offsets[0] = ((uint32*)(*slot)->values)[1];
                
                if (dir->parent_dir &&
                    dir->parent_dir->num_children > dir->parent_dir->cur_num_children+1 &&
                    dir->next_offset)
                  dir->parent_dir->child_offsets[dir->parent_dir->cur_num_children+1] = dir->next_offset; 
                
                reached_end_p = 1;
                    
                /* ### ? */
                break;

              case GFFTAG_NUM_CHILDREN:
                assert((*slot)->values);
                dir->num_children = ((uint32*)(*slot)->values)[0];
                /* ### FIXME: dynamic alloc */
#if 1
                dir->child_offsets = (uint32*)calloc(1024, sizeof(uint32));
                dir->child_dirs = (gff_dir**)calloc(1024, sizeof(gff_dir*));
#else
                dir->child_offsets = (uint32*)calloc(dir->num_children, sizeof(uint32));
                dir->child_dirs = (gff_dir**)calloc(dir->num_children, sizeof(gff_dir*));
#endif
                break;

              default:
                if ((*slot)->type == GFFTYPE_LDB)
                  { /* LDB follows */
                    data_offset = (*slot)->offset;
                  }
                else
                  { /* any other tag */
                    /* Anything to do? */
                  }
              }
          }
      }
          
      /* Read field data -- except trailing GFFTAG_DATA field. */
      {
        gff_tag  *tag;
        for (tag=dir->tags; tag; tag=tag->cdr)
          if (!tag->values && tag->type != GFFTYPE_LDB)
            {
              if (lseek(file->fd, tag->offset, SEEK_SET) >= 0)
                {
                  int32  size = tag->length * gff_type_size[tag->type];
                  tag->values = malloc(10*size);
                  if (gff_read(file, tag->values, size) != size)
                    {
                      fprintf(stderr, "%s:%d:fetch_dir:read: %s\n",
                              __FILE__, __LINE__, strerror(errno));
                      free_dir(dir);
                      return NULL; /* Memory leak? */
                    }
                  swap_data(tag->type, tag->length, tag->values);
                }
              else
                {
                  SEEK_ERROR();
                  free_dir(dir);
                  return NULL; /* Memory leak? */
                }
            }
      }

#ifdef GFF_DEBUG_OUTPUT
      /* Dump tags. */
      {
        gff_tag  *tag;
        for (tag=dir->tags; tag; tag=tag->cdr)
          dump_tag(stderr, tag, gff_get_tag_string);
      }
#endif

      /* Finally, determine next directory to read; eventually seek to data block. */
      {
        if (dir->child_offsets)
          {
            file->next_dir_to_read_offs = dir->child_offsets[0];
            file->next_dir_to_read_parent = dir;
          }
        else if (dir->up_link_offset)
          {
            assert(dir->parent_dir);
            file->next_dir_to_read_offs = dir->up_link_offset;
            file->next_dir_to_read_parent = dir->parent_dir->parent_dir;
          }
        else
          {
            file->next_dir_to_read_offs = dir->next_offset;
            file->next_dir_to_read_parent = dir->parent_dir;
          }
        
        if (data_offset)
          if (lseek(file->fd, data_offset, SEEK_SET) < 0)
            {
              SEEK_ERROR();
              free_dir(dir);
              return NULL;
            }
      }

      /* Successfully read directory. */
      file->currently_read_dir = dir;
      if (dir->parent_dir)
        {
          /* ### dynamic alloc! */
          assert((unsigned)dir->parent_dir->num_children > (unsigned)dir->parent_dir->cur_num_children);
          dir->parent_dir->child_offsets[dir->parent_dir->cur_num_children] = dir->offset; 
          dir->parent_dir->child_dirs[dir->parent_dir->cur_num_children] = dir;
          dir->parent_dir->cur_num_children++;
        }
      return file->currently_read_dir;
    }
  else
    {
      HEAP_OVERFLOW();
      return NULL;
    }
}

static gff_dir  *find_dir_at_already_read_offs(gff_dir *dir_tree, uint32 offset)
{
  if (!dir_tree)
    return NULL;
  else if (dir_tree->offset == offset)
    return dir_tree;
  else
    {
      gff_dir  *result = NULL;
      int  i;
      for (i=0; i<dir_tree->num_children; i++)
        if (dir_tree->child_offsets[i] == offset &&
            dir_tree->child_dirs[i])
          return dir_tree->child_dirs[i];
      if (dir_tree->next_dir)
        result = find_dir_at_already_read_offs(dir_tree->next_dir, offset);
      else if (dir_tree->up_link_dir)
        result = find_dir_at_already_read_offs(dir_tree->up_link_dir, offset);
      return result;
    }
}

/*! Reads subsequent directories, until directory at \a offs is
 *  reached.
 */
static gff_dir  *seek_dir_search(gff_file *file, uint32 offs)
{
  gff_dir  *dir;
  
  while (file->next_dir_to_read_offs)
    {
      if (lseek(file->fd, file->next_dir_to_read_offs, SEEK_SET) < 0)
        {
          fprintf(stderr, "%s:%d: could not seek to next directory\n",
                  __FILE__, __LINE__);
          return NULL;
        }
      dir = fetch_dir(file, file->next_dir_to_read_parent);

      if (!dir)
        return NULL;

      if (dir->tags->id == GFFTAG_UPLINK)
        {
          if (dir->parent_dir)
            {
              if (dir->parent_dir->num_children != GFF_UNKNOWN_SIZE &&
                  dir->parent_dir->num_children != dir->parent_dir->cur_num_children)
                fprintf(stderr, "%s:%d: WARNING: Child count %d doesn't fit the expected (%d).\n",
                        __FILE__, __LINE__, dir->parent_dir->cur_num_children-1, dir->parent_dir->num_children-1);
              dir->parent_dir->num_children = dir->parent_dir->cur_num_children;
            }
        }

      if (dir->offset == offs)
        return dir;
    }
  return NULL;
}

/*! Sets seeked_dir as the active directory. If \a seeked_dir is \e
 *  NULL, the directory at offs is seeked and it's new gff_dir struct
 *  is made active.
 *
 *  Note: The directory is \e not just seeked by setting the file
 *  pointer to \a offs. Instead, all directories between the actual
 *  position and \a offs are read, as well.
 *
 *    \param file       Current file.
 *    \param seeked_dir Directory, that should be seeked; may be \e NULL.
 *    \param offs       If seeked_dir is \e NULL: File offset of the seeked
 *                      directory.
 *
 *    \return           Pointer to the active directory; \e NULL on failure.
 */
static gff_dir  *seek_dir(gff_file *file, gff_dir *seeked_dir, uint32 offs)
{
#if 0
  fprintf(stderr, "%s:%d: >>>> seek for 0x%08x at 0x%08x\n",
          __FILE__, __LINE__, (unsigned)seeked_dir, (unsigned)offs);
#endif
  /* If seeked_dir is NULL: First have a look, whether the directory
     at offs has been read already... */
  if (!seeked_dir)
    seeked_dir = find_dir_at_already_read_offs(file->root_dir, offs);

  /* If seeked_dir is NULL and the directory hasn't been read yet:
     Read subsequent directories, until directory at offs is
     reached. */
  if (!seeked_dir)
    seeked_dir = seek_dir_search(file, offs);

  if (seeked_dir)
    /* Found the directory. Set it as active. */
    file->active_dir = seeked_dir;
  else
    /* couldn't find dir */
    file->active_dir = NULL;

  return file->active_dir;
}

gff_directory  *gff_read_root_dir(gff_file *file)
{
  return file->root_dir = seek_dir(file, file->root_dir, file->hd.root);
}

gff_directory  *gff_read_next_dir(gff_file *file, gff_directory *cur_dir)
{
  return D(cur_dir)->next_dir = seek_dir(file,
                                         D(cur_dir)->next_dir,
                                         D(cur_dir)->next_offset);
}

gff_directory  *gff_read_child_dir(gff_file *file, gff_directory *cur_dir, int child_idx)
{
  if (child_idx >= 0 && child_idx < D(cur_dir)->num_children-1) /* n-1 visible children */
    {
      /* ### dynamic allocation */

      if (!D(cur_dir)->child_dirs[child_idx] &&
          !D(cur_dir)->child_offsets[child_idx])
        gff_read_child_dir(file, cur_dir, child_idx-1);

      D(cur_dir)->child_dirs[child_idx] =
        seek_dir(file, D(cur_dir)->child_dirs[child_idx], D(cur_dir)->child_offsets[child_idx]);
      if (D(cur_dir)->child_dirs[child_idx])
        {
          if (D(cur_dir)->child_dirs[child_idx]->tags->id == GFFTAG_UPLINK)
            {
              NEVER_REACHED();
              return NULL;
            }
          else
            return D(cur_dir)->child_dirs[child_idx];
        }
      else
        /* Must have been a wrong child count. seek_dir() should
           already have printed a warning. */
        return NULL;
    }
  else
    {
      fprintf(stderr, "%s:%d:gff_read_child_dir: child index %d out of bounds\n",
              __FILE__, __LINE__, child_idx);
      return NULL;
    }
}

uint32  gff_dir_get_num_children(const gff_directory *dir)
{
  if (cD(dir)->num_children == GFF_UNKNOWN_SIZE)
    return GFF_UNKNOWN_SIZE;
  else
    return cD(dir)->num_children-1; /* n-1 visible children */
}

uint32  gff_dir_get_num_tags(const gff_directory *dir)
{
  return cD(dir)->num_tags;
}

/*! The tag id's of all tags in \a dir are written to \a tag_ids. \a
 *  tag_ids must provide sufficient space. Use gff_dir_get_num_tags()
 *  to determine the actual number of tags in \a dir.
 *
 *    \return           0 on success
 */
int  gff_dir_get_tag_list(const gff_directory *dir, int *tag_ids)
{
  int             i;
  const gff_tag  *t;
  for (i=0, t=cD(dir)->tags; i<cD(dir)->num_tags && t; t=t->cdr, i++)
    tag_ids[i] = t->id;
  assert(!t);
  assert(i<cD(dir)->num_tags);
  return 0;
}

const gff_directory  *gff_dir_get_parent_dir(const gff_directory *dir)
{
  return cD(dir)->parent_dir;
}

int  gff_query_field(const gff_directory *dir, int tag_id, int *field_type, int *num_elements)
{
  const gff_tag  *t = find_tag(cD(dir), tag_id);
  if (t)
    {
      if (field_type)    *field_type = t->type;
      if (num_elements)  *num_elements = t->length;
      return 0;
    }
  else
    return 1;
}

int  gff_get_field(const gff_directory *dir, int tag_id, int field_type, int expected_size, ...)
{
  int      result;
  va_list  ap;
  
  va_start(ap, expected_size);
  result = gff_v_get_field(dir, tag_id, field_type, expected_size, ap);
  va_end(ap);
  
  return result;
}

/*!   \return           Number of tag values; < 0 on error.\
 */
int  gff_v_get_field(const gff_directory *dir, int tag_id, int field_type, int expected_size, va_list ap)
{
  const gff_tag  *t = find_tag(cD(dir), tag_id);
  if (t && t->type == field_type)
    {
      int  i, n = (int)t->length < expected_size ? (int)t->length : expected_size;
      switch (t->type)
        {
        case GFFTYPE_UINT8:
        case GFFTYPE_INT8:
          for (i=0; i<n; i++)
            {
              uint8  *ptr = va_arg(ap, uint8*);
              *ptr = ((const uint8*)t->values)[i];
            }
          break;
        case GFFTYPE_UINT16:
        case GFFTYPE_INT16:
          for (i=0; i<n; i++)
            {
              uint16  *ptr = va_arg(ap, uint16*);
              *ptr = ((const uint16*)t->values)[i];
            }
          break;
        case GFFTYPE_UINT32:
        case GFFTYPE_INT32:
          for (i=0; i<n; i++)
            {
              uint32  *ptr = va_arg(ap, uint32*);
              *ptr = ((const uint32*)t->values)[i];
            }
          break;
        case GFFTYPE_FLOAT:
          for (i=0; i<n; i++)
            {
              float  *ptr = va_arg(ap, float*);
              *ptr = ((const float*)t->values)[i];
            }
          break;
        case GFFTYPE_DOUBLE:
          for (i=0; i<n; i++)
            {
              double  *ptr = va_arg(ap, double*);
              *ptr = ((const double*)t->values)[i];
            }
          break;
        case GFFTYPE_ASCII:
          {
            char  *ptr = va_arg(ap, char*);
            memcpy(ptr, t->values, n);
          }
          break;
        default:
          fprintf (stderr, "%s:%d:gff_v_get_field: You're not supposed to read tags of type #%d!\n",
                   __FILE__, __LINE__, field_type);
        }
      return t->length;
    }
  return -1;
}

int  gff_get_field_arr(const gff_directory *dir, int tag_id, int field_type, int expected_size, void *dest)
{
  const gff_tag  *t = find_tag(cD(dir), tag_id);
  if (t)
    if (t->type == field_type)
      {
        int  n = (int)t->length < expected_size ? (int)t->length : expected_size;
        memcpy(dest, t->values, n*gff_type_size[field_type]);
        return n;
      }
  return -1;
}

const void  *gff_get_field_arr_p(const gff_directory *dir, int tag_id, int field_type)
{
  const gff_tag  *t = find_tag(cD(dir), tag_id);
  if (t)
    if (t->type == field_type)
      return t->values;
  return NULL;
}

int  gff_seek_field_data(const gff_file *file, const gff_directory *dir, int tag_id)
{
  if (file->seekable_p)
    {
      const gff_tag  *t = find_tag(cD(dir), tag_id);
      if (t)
        if (lseek(file->fd, t->offset, SEEK_SET) >= 0)
          return 0;
    }
  return 1;
}


/*****************************************************************************
 *  Write mode
 *****************************************************************************/
  
static int  v_write_field(gff_file *file, int tag_id, int field_type, int num_elements, va_list ap)
{
  gff_tag  *t;
  if (!create_tag(file->active_dir, tag_id, field_type, num_elements, &t))
    if (t)
      {
        int  i, n = t->length = num_elements;
        int  ok = 1;
        switch (t->type)
          {
          case GFFTYPE_UINT8:
          case GFFTYPE_INT8:
            for (i=0; i<n; i++)
              ((uint8*)t->values)[i] = va_arg(ap, int);
            break;
          case GFFTYPE_UINT16:
          case GFFTYPE_INT16:
            for (i=0; i<n; i++)
              ((uint16*)t->values)[i] = va_arg(ap, int);
            break;
          case GFFTYPE_UINT32:
          case GFFTYPE_INT32:
            for (i=0; i<n; i++)
              ((uint32*)t->values)[i] = va_arg(ap, uint32);
            break;
          case GFFTYPE_FLOAT:
            /* Note, that float values are passed as doubles to va_list! */
            for (i=0; i<n; i++)
              ((float*)t->values)[i] = (float)va_arg(ap, double);
            break;
          case GFFTYPE_DOUBLE:
            for (i=0; i<n; i++)
              ((double*)t->values)[i] = va_arg(ap, double);
            break;
          case GFFTYPE_ASCII:
            memcpy(t->values, va_arg(ap, const char*), n);
            break;
          default:
            fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write tags of type #%d!\n",
                     __FILE__, __LINE__, field_type);
            ok = 0;
          }
        return !ok;
      }
    else
      HEAP_OVERFLOW();
  else
    fprintf(stderr, "%s:%d:gff_write_field: tag 0x%04x already exists\n",
            __FILE__, __LINE__, (unsigned)tag_id);
  return 1;
}

int  gff_write_field(gff_file *file, int tag_id, int field_type, int num_elements, ...)
{
  int      result;
  va_list  ap;
  
  va_start(ap, num_elements);
  result = gff_v_write_field(file, tag_id, field_type, num_elements, ap);
  va_end(ap);
  
  return result;
}

int  gff_v_write_field(gff_file *file, int tag_id, int field_type, int num_elements, va_list ap)
{
  if (field_type != GFFTYPE_LDB && field_type != GFFTYPE_OFFSET && field_type != GFFTYPE_2OFFSET)
    if (tag_id >= GFF_INTERNAL_MAXTAG)
      return v_write_field(file, tag_id, field_type, num_elements, ap);
    else
      fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write internal GFF tags!\n",
               __FILE__, __LINE__);
  else
    fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write tags of type #%d!\n",
             __FILE__, __LINE__, field_type);
  return 1;
}

int  gff_write_field_arr(gff_file *file, int tag_id, int field_type, int num_elements, const void *data)
{
  if (field_type != GFFTYPE_LDB && field_type != GFFTYPE_OFFSET && field_type != GFFTYPE_2OFFSET)
    if (tag_id >= GFF_INTERNAL_MAXTAG)
      if (data)
        if (num_elements != GFF_UNKNOWN_SIZE)
          {
            gff_tag  *t;
            if (!create_tag(file->active_dir, tag_id, field_type, num_elements, &t))
              if (t)
                {
                  memcpy(t->values, data, t->length*gff_type_size[t->type]);
                  return 0;
                }
              else
                HEAP_OVERFLOW();
            else
              fprintf(stderr, "%s:%d:gff_write_field: tag 0x%04x already exists\n",
                      __FILE__, __LINE__, (unsigned)tag_id);
          }
        else
          fprintf(stderr, "%s:%d:gff_write_field: Use gff_write_data_field(), if the array size is unknown.\n",
                  __FILE__, __LINE__);
      else
        fprintf(stderr, "%s:%d:gff_write_field: You must specify the data pointer.\n",
                __FILE__, __LINE__);
    else
      fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write internal GFF tags!\n",
               __FILE__, __LINE__);
  else
    fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write tags of type #%d!\n",
             __FILE__, __LINE__, field_type);
  return 1;
}

int  gff_write_field_cstring(gff_file *file, int tag_id, const char *str)
{
  return gff_write_field_arr(file, tag_id, GFFTYPE_ASCII, strlen(str)+1, str);
}

/*****************************************************************************/

/*! \internal
 *
 *  A function, that calles break_dir() must provide a new allocated
 *  gff_dir struct in \a next_dir, that will receive the next written
 *  directory. However, the caller must also fill all
 *  next_dir/child_dirs/up_link_dir entries of \a next_dir's
 *  predecessors!
 *
 *  Note, that next_dir is NULL, it the very last directory is
 *  written.
 */
static int  break_dir(gff_file *file, gff_dir *next_dir, int eof_tag_p)
{
  gff_dir  *dir = file->active_dir;
  gff_tag  *tag;

  int  i;

  uint32  tail_offs;

  if (dir->parent_dir)
    dir->parent_dir->cur_num_children++;

  for (tag=dir->tags, dir->num_tags=0; tag; dir->num_tags++, tag=tag->cdr);

  dir->offset = lseek(file->fd, 0, SEEK_CUR);
  if (GFF_FILE_PADDING_FRAC(dir->offset))
    {
      int   m = GFF_FILE_PADDING_CEIL(dir->offset) - dir->offset;
      char  zeroes[GFF_FILE_PADDING] = {0};
      if (gff_write(file, &zeroes[0], m) != m)
        {
          WRITE_ERROR();
          return 1;
        }
      dir->offset += m;
    }
  assert(GFF_FILE_PADDING_FRAC(dir->offset) == 0);
  assert(GFF_FILE_PADDING_FRAC(lseek(file->fd, 0, SEEK_CUR)) == 0);

  tail_offs = dir->offset + GFF_FILE_PADDING_CEIL(dir->num_tags * sizeof(gff_tag_on_disk));

  {
    /* Allocate n+1, so we are able to easily write more bytes to preserve padding: */
    gff_tag_on_disk  *phystags = (gff_tag_on_disk *)alloca((dir->num_tags+1) * sizeof(*phystags));
    int               saw_ldb_p = 0;

    /* Collect tags into write-buffer */
    for (tag=dir->tags, i=0; tag; i++, tag=tag->cdr)
      {
        tag->target_offset = dir->offset + i*sizeof(gff_tag_on_disk);

        switch (tag->id)
          {
          case GFFTAG_UPLINK:
            assert(!tag->cdr);  /* must be the last tag in dir */
            assert(eof_tag_p || dir->parent_dir); /* must be a child */

            if (dir->parent_dir)
              ((uint32 *)tag->values)[0] = tail_offs;
            else
              ((uint32 *)tag->values)[0] = 0;

            /* if file is seekable, update some directory information */
            if (file->seekable_p && !eof_tag_p)
              {
                if (dir->parent_dir->num_children == GFF_UNKNOWN_SIZE ||
                    dir->parent_dir->cur_num_children != dir->parent_dir->num_children)
                  {
                    /* update the parent's num_children entry */
                    if (dir->parent_dir->num_children != GFF_UNKNOWN_SIZE)
                      {
                        fprintf(stderr, "%s:%d: WARNING: Children count %d differs from the announced (%d).\n",
                                __FILE__, __LINE__,
                                dir->parent_dir->cur_num_children-1, dir->parent_dir->num_children-1);
                      }
                    dir->parent_dir->num_children = dir->parent_dir->cur_num_children;
                    if (lseek(file->fd, dir->parent_dir->num_children_offset, SEEK_SET) < 0)
                      {
                        SEEK_ERROR();
                        return 1;
                      }
                    {
                      uint32  num = dir->parent_dir->num_children;
                      pi_uint32_to_net_ip(&num);
                      gff_write(file, &num, sizeof(num));
                    }
                  }

                {
                  /* update the parent's next entry */
                  if (lseek(file->fd, dir->parent_dir->next_offset_offset, SEEK_SET) < 0)
                    {
                      SEEK_ERROR();
                      return 1;
                    }
                  {
                    uint32  offs = tail_offs;
                    pi_uint32_to_net_ip(&offs);
                    gff_write(file, &offs, sizeof(offs));
                  }
                }

                /* seek back, where you came from */
                if (lseek(file->fd, dir->offset, SEEK_SET) < 0)
                  {
                    SEEK_ERROR();
                    return 1;
                  }
              }
            break;

          case GFFTAG_LINK:
            assert (!tag->cdr); /* must be the last tag in dir */
            ((uint32 *)tag->values)[0] = tail_offs;
            dir->next_offset_offset = tag->target_offset + (((uint8*)tag->values) - (uint8*)&tag->id);
            break;

          case GFFTAG_NUM_CHILDREN:
#if 1
            ((uint32 *)tag->values)[0] = dir->num_children;
#else
            ((uint32 *)tag->values)[0] = GFF_UNKNOWN_SIZE;
#endif
            dir->num_children_offset = tag->target_offset + (((uint8*)tag->values) - (uint8*)&tag->id);
            break;

          case GFFTAG_DOWNLINK:
            assert (!tag->cdr); /* must be the last tag in dir */
            ((uint32 *)tag->values)[0] = 0;
            ((uint32 *)tag->values)[1] = tail_offs;
            dir->next_offset_offset = tag->target_offset + (((uint8*)tag->values) - (uint8*)&tag->id);
            break;

          default:
            if (!TAG_IS_INLINED(tag) && tag->length) /* This includes tag->length == GFF_UNKNOWN_SIZE! */
              tag->offset = tail_offs;

            if (tag->type == GFFTYPE_LDB)
              {
                if (!saw_ldb_p)
                  saw_ldb_p = 1;
                else
                  {
                    NEVER_REACHED();
                    return 1;
                  }
              }
            else if (saw_ldb_p)
              {
                NEVER_REACHED();
                return 1;
              }
          }

#ifdef GFF_DEBUG_OUTPUT
        dump_tag(stderr, tag, gff_get_tag_string);
#endif

        /* prepare tag for disk */
        {
          phystags[i].id = tag->id;
          phystags[i].type = tag->type;
          phystags[i].length = tag->length;
          phystags[i].offset = tag->offset;

          pi_uint16_to_net_ip(&phystags[i].id);
          pi_uint16_to_net_ip(&phystags[i].type);
          pi_uint32_to_net_ip(&phystags[i].length);

          if (tag->type == GFFTYPE_2OFFSET || !TAG_IS_INLINED(tag))
            {
              /* GFFTYPE_2OFFSET or non-inlined data */
              pi_uint32_to_net_ip(&phystags[i].offset);
            }
          else
            {
              /* data is inlined in offset-field */
              assert(tag->values == &tag->offset);

              /* swap inlined data */
              swap_data(tag->type, tag->length, &phystags[i].offset);
            }
        }

        /* eventually shift tail_offs */
        if (!TAG_IS_INLINED(tag))
          {
            if (tag->length != GFF_UNKNOWN_SIZE)
              tail_offs += GFF_FILE_PADDING_CEIL(tag->length * gff_type_size[tag->type]);
            else if (tag->type == GFFTYPE_LDB)
              tail_offs = GFF_UNKNOWN_SIZE;
            else
              {
                NEVER_REACHED();
                return 1;
              }
          }
      }

    /* Dump tags from write-buffer. */
    if (gff_write(file, &phystags[0], GFF_FILE_PADDING_CEIL(dir->num_tags * sizeof(phystags[0]))) < 0)
      {
        WRITE_ERROR();
        return 1;
      }

    /* Dump non-inlined tag data. */
    for (tag=dir->tags, i=0; tag; i++, tag=tag->cdr)
      if (!TAG_IS_INLINED(tag) && tag->length)
        {
          assert(!GFF_FILE_PADDING_FRAC(tail_offs));
          if (tag->type != GFFTYPE_LDB)
            {
              assert(tag->values);
              assert(tag->length != GFF_UNKNOWN_SIZE);

              /* swap tag data */
              swap_data(tag->type, tag->length, tag->values);

              /* write tag data */
              {
                int   n = tag->length * gff_type_size[tag->type];
                int   m = GFF_FILE_PADDING_CEIL(n) - n;
                if (gff_write(file, tag->values, n) != n)
                  {
                    WRITE_ERROR();
                    return 1;
                  }
                if (m)
                  {
                    char  zeroes[GFF_FILE_PADDING] = {0};
                    if (gff_write(file, &zeroes[0], m) != m)
                      {
                        WRITE_ERROR();
                        return 1;
                      }
                  }
              }

              /* free tag data */
              free (tag->values);
            }
          else
            { /* This tag starts an LDB */
            }
        }
    assert(saw_ldb_p || lseek(file->fd, 0, SEEK_CUR) == (ssize_t)tail_offs);
    assert(i == dir->num_tags);
    /*fprintf(stderr, "%08x vs %08x\n", (unsigned)lseek(file->fd, 0, SEEK_CUR), (unsigned)tail_offs);*/
  }

  file->active_dir = next_dir;

  return 0;
}

/*! Set num_children to GFF_UNKNOWN_SIZE, if the number of children is still unknown.
 */
int  gff_write_downlink(gff_file *file, uint32 num_children)
{
  gff_tag  *t;
  if (!create_tag(file->active_dir, GFFTAG_NUM_CHILDREN, GFFTYPE_UINT32, 1, &t))
    if (t)
      if (!create_tag(file->active_dir, GFFTAG_DOWNLINK, GFFTYPE_2OFFSET, 1, &t))
        if (t)
          {
            gff_dir  *first_child;

            if (num_children != GFF_UNKNOWN_SIZE)  ++num_children; /* including uplink */
            file->active_dir->num_children = num_children;
#if 1
            /* ### FIXME: dynamic alloc */
            file->active_dir->child_dirs = (gff_dir **)calloc(1024, sizeof(gff_dir*));
            file->active_dir->child_offsets = (uint32 *)calloc(1024, sizeof(uint32));
#else
            file->active_dir->child_dirs = (gff_dir **)calloc(num_children, sizeof(gff_dir*));
            file->active_dir->child_offsets = (uint32 *)calloc(num_children, sizeof(uint32));
#endif

            first_child = new_dir();
            file->active_dir->child_dirs[0] = first_child;
            first_child->parent_dir = file->active_dir;

            return break_dir(file, first_child, 0);
          }
        else
          HEAP_OVERFLOW();
      else
        NEVER_REACHED();
    else
      HEAP_OVERFLOW();
  else
    NEVER_REACHED();
  return 1;
}

static int  write_uplink(gff_file *file, int eof_tag_p)
{
  if (file->active_dir->parent_dir || eof_tag_p)
    {
      gff_tag  *t;

      if (file->active_dir->num_tags)
        /* We only want a sole uplink node */
        if (gff_break_dir(file))
          return 1;

      if (!create_tag(file->active_dir, GFFTAG_UPLINK, GFFTYPE_OFFSET, 1, &t))
        if (t)
          {
            if (!eof_tag_p)
              {
                gff_dir  *next_dir = new_dir();
                assert(file->active_dir->parent_dir);
                if (next_dir)
                  {
                    if (file->active_dir->parent_dir->num_children != GFF_UNKNOWN_SIZE &&
                        (file->active_dir->parent_dir->num_children
                         != file->active_dir->parent_dir->cur_num_children+1))
                      {
                        fprintf(stderr, "%s:%d: WARNING: Children count %d differs from the announced (%d)!!\n",
                                __FILE__, __LINE__,
                                file->active_dir->parent_dir->cur_num_children+1-1,
                                file->active_dir->parent_dir->num_children-1);
                      }
                    /* Don't update
                       file->active_dir->parent_dir->num_children, so
                       break_dir knows, that the entry must be
                       updated */
                    /*file->active_dir->parent_dir->num_children = file->active_dir->parent_dir->cur_num_children+1;*/

                    file->active_dir->up_link_dir = next_dir;

                    next_dir->parent_dir = file->active_dir->parent_dir->parent_dir;
            
                    return break_dir(file, next_dir, 0);
                  }
                else
                  HEAP_OVERFLOW();
              }
            else
              return break_dir(file, NULL, 1);
          }
        else
          HEAP_OVERFLOW();
      else
        NEVER_REACHED();
    }
  else
    fprintf(stderr, "%s:%d: Can not insert up-link on top-level.\n",
            __FILE__, __LINE__);
  return 1;
}

int  gff_write_uplink(gff_file *file)
{
  return write_uplink(file, 0);
}

int  gff_write_data_field(gff_file *file, int tag_id, int num_bytes)
{
  if (tag_id >= GFF_INTERNAL_MAXTAG)
    {
      gff_tag  *data_tag;

      if (!create_tag(file->active_dir, tag_id, GFFTYPE_LDB, 0, &data_tag))
        if (data_tag)
          {
            /* This tag is meant to describe a following large data block. */
            assert(!data_tag->values);
            data_tag->length = num_bytes;
            return gff_break_dir(file); /* This also produces a GFFTAG_LINK tag. */
          }
        else
          HEAP_OVERFLOW();
      else
        NEVER_REACHED();
    }
  else
    fprintf (stderr, "%s:%d:gff_write_field: You're not supposed to write internal GFF tags!\n",
             __FILE__, __LINE__);
  return 1;
}

int  gff_break_dir(gff_file *file)
{
  gff_tag  *link_tag;

  if (!create_tag(file->active_dir, GFFTAG_LINK, GFFTYPE_OFFSET, 1, &link_tag))
    if (link_tag)
      {
        gff_dir  *next_dir = new_dir();
        if (next_dir)
          {
            next_dir->parent_dir = file->active_dir->parent_dir;
            
            file->active_dir->next_dir = next_dir;
            /* (FIXME: PROVE:) next_offset will be written-back in break_dir automatically */
            
            return break_dir(file, next_dir, 0);
          }
        else
          HEAP_OVERFLOW();
      }
    else
      HEAP_OVERFLOW();
  else
    NEVER_REACHED();
  return 1;
}

/* Some Emacs-Hints -- please don't remove:
 *
 *  Local Variables:
 *  tab-width:4
 *  End:
 */
