// ### CHECK-ME-BEFORE-RELEASE ###
//
// Title:   instream.cpp
// Created: Thu Apr 18 14:30:59 2002
// Authors: Tim Weyrich <weyrich@inf.ethz.ch>
//
// Copyright (c) 2001--2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//
// ---------------------------------------------------------------
//
// $Log: instream.cpp,v $
// Revision 1.12  2004/04/27 15:34:06  weyrich
// experimental: fixing propmask/layout inconsistencies caused by the Pointshop3D writer
//
// Revision 1.11  2003/11/19 13:56:48  weyrich
// fix: property read
//
// Revision 1.10  2003/11/19 13:04:01  weyrich
// added Pointshop3D headers
//
// Revision 1.9  2003/11/17 15:05:52  weyrich
// new API allowing for more than 32 surfel properties
//
// Revision 1.8  2003/10/09 15:31:54  wuermlin
// removed _init and _fini
//
// Revision 1.7  2003/03/20 14:52:51  weyrich
// smaller fixes of implicit coordinate system transformations
//
// Revision 1.6  2002/12/18 17:05:45  weyrich
// added feature relations support (experimental)
//
// Revision 1.5  2002/11/26 16:17:24  weyrich
// moved _init/_fini definition in instream.cpp
//
// Revision 1.4  2002/11/18 13:09:12  weyrich
// added markers and tangential vectors
//
// Revision 1.3  2002/05/23 13:59:56  weyrich
// First official release.
//
// Revision 1.2  2002/05/17 13:58:16  weyrich
// PointShop3D integration nearly finished
//
// Revision 1.1.1.1  2002/05/08 08:58:47  weyrich
// surfel file format
//

#include <pisystem.h>  // for alloca
#include <pistring.h>
#include <pimath.h>

#include "api.h"
#include "internal/sflmat.h"

#if 0
extern "C" void _init(void)
{
}

extern "C" void _fini(void)
{
}
#endif

using namespace sfl;
using namespace sflMat;

#define  HEAP_OVERFLOW()                                                       \
do { fprintf(stderr, "%s:%d: HEAP OVERFLOW\n", __FILE__, __LINE__); } while(0)

///////////////////////////////////////////////////////////////////////////////

InStream::InStream()
  : readWorldFlag(false)
{
  setIdentity32(fileCoordSystem);
  setIdentity32(appCoordSystem);
}

InStream::~InStream()
{
}

///////////////////////////////////////////////////////////////////////////////

char   *InStream::getFieldString(gff_directory *dir, int tagId, const char *deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_ASCII)
      {
        char  *res = (char *)malloc(size);
        if (res)
          {
            if (gff_get_field_arr(dir, tagId, GFFTYPE_ASCII, size, res) >= 0)
              return res;
            free(res);
          }
        else
          HEAP_OVERFLOW();
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be ASCII, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (deflt)
    return strdup(deflt);
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return NULL;
}

bool    InStream::getFieldString(gff_directory *dir, int tagId, char *dest, size_t buflen, const char *deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_ASCII)
      {
        if ((int)buflen > size)
          buflen = size;
        if (gff_get_field_arr(dir, tagId, GFFTYPE_ASCII, buflen, dest) >= 0)
          {
            dest[buflen-1] = 0;
            return true;
          }
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be ASCII, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (deflt)
    {
      size = strlen(deflt)+1;
      if ((int)buflen > size)
        buflen = size;
      memcpy(dest, deflt, buflen);
      dest[buflen-1] = 0;
      return true;
    }
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return false;
}

bool    InStream::getFieldUInt8(gff_directory *dir, int tagId, uint8 &val, bool haveDefault, uint8 deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_UINT8 && size == 1)
      {
        if (gff_get_field_arr(dir, tagId, GFFTYPE_UINT8, 1, &val) == 1)
          return true;
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be a UINT8, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    {
      val = deflt;
      return true;
    }
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return false;
}

uint8  InStream::getFieldUInt8(gff_directory *dir, int tagId, bool haveDefault, uint8 deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_UINT8 && size == 1)
      {
        uint8  res;
        if (gff_get_field(dir, tagId, GFFTYPE_UINT8, 1, &res) == 1)
          return res;
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be UINT8, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    return deflt;
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return 0;
}

bool    InStream::getFieldUInt32(gff_directory *dir, int tagId, uint32 &val, bool haveDefault, uint32 deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_UINT32 && size == 1)
      {
        if (gff_get_field_arr(dir, tagId, GFFTYPE_UINT32, 1, &val) == 1)
          return true;
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be a UINT32, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    {
      val = deflt;
      return true;
    }
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return false;
}

uint32  InStream::getFieldUInt32(gff_directory *dir, int tagId, bool haveDefault, uint32 deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if (type == GFFTYPE_UINT32 && size == 1)
      {
        uint32  res;
        if (gff_get_field(dir, tagId, GFFTYPE_UINT32, 1, &res) == 1)
          return res;
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be UINT32, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    return deflt;
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return 0;
}

bool    InStream::getFieldDouble(gff_directory *dir, int tagId, double &val, bool haveDefault, double deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if ((type == GFFTYPE_DOUBLE || type == GFFTYPE_FLOAT) && size == 1)
      {
        if (type == GFFTYPE_DOUBLE)
          {
            if (gff_get_field_arr(dir, tagId, GFFTYPE_DOUBLE, 1, &val) == 1)
              return true;
          }
        else
          {
            float  temp;
            if (gff_get_field_arr(dir, tagId, GFFTYPE_FLOAT, 1, &temp) == 1)
              {
                val = temp;
                return true;
              }
          }
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be a DOUBLE, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    {
      val = deflt;
      return true;
    }
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return false;
}

double  InStream::getFieldDouble(gff_directory *dir, int tagId, bool haveDefault, double deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    if ((type == GFFTYPE_DOUBLE || type == GFFTYPE_FLOAT) && size == 1)
      {
        if (type == GFFTYPE_DOUBLE)
          {
            double  res;
            if (gff_get_field_arr(dir, tagId, GFFTYPE_DOUBLE, 1, &res) == 1)
              return res;
          }
        else
          {
            float  res;
            if (gff_get_field_arr(dir, tagId, GFFTYPE_FLOAT, 1, &res) == 1)
              return res;
          }
      }
    else
      fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be FLOAT or DOUBLE, but turned out to be not...\n",
              __FILE__, __LINE__, (unsigned)tagId);
  else if (haveDefault)
    return deflt;
  else
    fprintf(stderr, "%s:%d: Field (id 0x%04x) does not exist.\n",
            __FILE__, __LINE__, (unsigned)tagId);
  return 0.0;
}

uint32  InStream::getFieldUInt32ArrIdx(gff_directory *dir, int tagId, int idx, uint32 deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    {
      if (type == GFFTYPE_UINT32)
        {
          if (idx >= 0 && idx < size)
            return ((uint32 *)gff_get_field_arr_p(dir, tagId, GFFTYPE_UINT32))[idx];
          else
            fprintf(stderr, "%s:%d: Field 0x%04x: index out of bounds 0..%d.\n",
                    __FILE__, __LINE__, (unsigned)tagId, size);
        }
      else
        fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be UINT32, but turned out to be not...\n",
                __FILE__, __LINE__, (unsigned)tagId);
    }
  return deflt;
}

double  InStream::getFieldDoubleArrIdx(gff_directory *dir, int tagId, int idx, double deflt)
{
  int  type;
  int  size;
  if (!gff_query_field(dir, tagId, &type, &size))
    {
      if (type == GFFTYPE_DOUBLE || type == GFFTYPE_FLOAT)
        {
          if (idx >= 0 && idx < size)
            {
              if (type == GFFTYPE_DOUBLE)
                return ((double *)gff_get_field_arr_p(dir, tagId, GFFTYPE_DOUBLE))[idx];
              else
                return ((float *)gff_get_field_arr_p(dir, tagId, GFFTYPE_FLOAT))[idx];
            }
          else
            fprintf(stderr, "%s:%d: Field 0x%04x: index out of bounds 0..%d.\n",
                    __FILE__, __LINE__, (unsigned)tagId, size);
        }
      else
        fprintf(stderr, "%s:%d: Field (id 0x%04x) is supposed to be FLOAT or DOUBLE, but turned out to be not...\n",
                __FILE__, __LINE__, (unsigned)tagId);
    }
  return deflt;
}

///////////////////////////////////////////////////////////////////////////////

InStream  *InStream::open(const char *fName, const char *mode)
{
  InStream  *result = new InStream();
  char  *m = (char *)alloca(strlen(mode)+2);

  strcpy(m, mode);
  strcat(m, "r");

  if ((result->gff = gff_open(fName, m, SFL_MAGIC, SFL_VERSION)))
    {
      if ((result->header = gff_read_root_dir(result->gff)))
        {
          if ((result->surfelSetList = gff_read_child_dir(result->gff, result->header,
                                                          getFieldUInt32(result->header,
                                                                         SFLTAG_SURFEL_SET_LIST_IDX,
                                                                         true, 0))))
            {
              // Read default units
              result->defaultUnits = getFieldDouble(result->header, SFLTAG_DEFAULT_UNITS, true, 1.0);

              // Read coordinate system
              {
                const float  *v;

                setIdentity16(result->fileCoordSystem);

                if ((v = (const float *)gff_get_field_arr_p(result->header, SFLTAG_RIGHT_VEC, GFFTYPE_FLOAT)))
                  {
                    result->fileCoordSystem[0] = v[0];
                    result->fileCoordSystem[4] = v[1];
                    result->fileCoordSystem[8] = v[2];
                  }

                if ((v = (const float *)gff_get_field_arr_p(result->header, SFLTAG_UP_VEC, GFFTYPE_FLOAT)))
                  {
                    result->fileCoordSystem[1] = v[0];
                    result->fileCoordSystem[5] = v[1];
                    result->fileCoordSystem[9] = v[2];
                  }

                if ((v = (const float *)gff_get_field_arr_p(result->header, SFLTAG_BACK_VEC, GFFTYPE_FLOAT)))
                  {
                    result->fileCoordSystem[2] = v[0];
                    result->fileCoordSystem[6] = v[1];
                    result->fileCoordSystem[10] = v[2];
                  }

                updateInverse(result->fileCoordSystem);
              }

              // Read camera matrix -- if there is one
              {
                const float  *M;
                if ((M = (const float *)gff_get_field_arr_p(result->header, SFLTAG_SCENE_CAMERA_TRANSFORM, GFFTYPE_FLOAT)))
                  {
                    int  i;
                    for (i=0; i<16; i++)
                      result->sceneCameraTransform[i] = M[i];
                    updateInverse(result->sceneCameraTransform);
                  }
                else
                  setIdentity32(result->sceneCameraTransform);
              }

              result->surfelSet = NULL;
              result->surfelSetIdx = -1;
              return result;
            }
          else
            fprintf(stderr, "%s:%d:sfl::InStream::open(): Could not read surfel set list directory.\n",
                    __FILE__, __LINE__);
        }
      else
        fprintf(stderr, "%s:%d:sfl::InStream::open(): Could not read root directory.\n",
                __FILE__, __LINE__);
      gff_close(result->gff);
    }

  delete result;
  return NULL;
}

int  InStream::close(InStream *in)
{
  delete in;
  return 0;
}

///////////////////////////////////////////////////////////////////////////////

InStream  &InStream::setLocalRightVector(double x, double y, double z)
{
  appCoordSystem[0] = x;
  appCoordSystem[4] = y;
  appCoordSystem[8] = z;
  return *this;
}

InStream  &InStream::setLocalUpVector(double x, double y, double z)
{
  appCoordSystem[1] = x;
  appCoordSystem[5] = y;
  appCoordSystem[9] = z;
  return *this;
}

InStream  &InStream::setLocalBackVector(double x, double y, double z)
{
  appCoordSystem[2] = x;
  appCoordSystem[6] = y;
  appCoordSystem[10] = z;
  return *this;
}

///////////////////////////////////////////////////////////////////////////////

bool  InStream::seekSurfelSet(int idx)
{
  if ((surfelSet = gff_read_child_dir(gff, surfelSetList, surfelSetIdx=idx)))
    {
      int            diskSize;
      int            numProps;
      const uint16  *propIdcs;
      const uint32  *propOffs;

      // Read surfel set format version.
      {
        surfelSetFormatVersion = getFieldUInt32(surfelSet, SFLTAG_SURFEL_SET_FORMAT_VERSION, true, 0);
        if ((surfelSetFormatVersion & 0xffff0000UL) >
            (SFL_SURFEL_FORMAT_VERSION & 0xffff0000UL))
          {
            fprintf(stderr, "sfl::InStream::seekSurfelSet(): Can not read surfel set format version 0x%08x.\n",
                    (unsigned)surfelSetFormatVersion);
            surfelSet = NULL;
            surfelSetIdx = -1;
            return false;
          }
      }

      // Read number of resolutions.
      surfelSetNumResolutions = getFieldUInt32(surfelSet,
                                               SFLTAG_SURFEL_SET_NUM_RESOLUTIONS,
                                               true, 1);

      // Read surfel set properties.
      {
        int  i;

        {
          int  n = queryFieldNumElements(surfelSet, SFLTAG_SURFEL_SET_PROPERTIES);
          uint32  *buf = new uint32[n];
          for (i=0; i<n; i++)
            buf[i] = getFieldUInt32ArrIdx(surfelSet, SFLTAG_SURFEL_SET_PROPERTIES, i, 0);

          surfelSetProperties = PropMaskType(32*n, buf);

          delete[] buf;
        }

        gff_query_field(surfelSet, SFLTAG_SURFEL_SET_PROP_INDICES, NULL, &numProps);
        propIdcs = (const uint16 *)gff_get_field_arr_p(surfelSet, SFLTAG_SURFEL_SET_PROP_INDICES, GFFTYPE_UINT16);
        propOffs = (const uint32 *)gff_get_field_arr_p(surfelSet, SFLTAG_SURFEL_SET_PROP_OFFSETS, GFFTYPE_UINT32);

        diskSize = getFieldUInt32(surfelSet, SFLTAG_SURFEL_SET_SURFEL_SIZE);
      }

      // Read surfel set units.
      surfelSetUnits = getFieldDouble(header, SFLTAG_SURFEL_SET_UNITS, true, defaultUnits);

      // Read surfel set transformation.
      {
        const float  *t = (const float *)gff_get_field_arr_p(surfelSet, SFLTAG_SURFEL_SET_TRANSFORM, GFFTYPE_FLOAT);
        bool          setId = true;

        if (t)
          {
            int  i;
            for (i=0; i<16; i++)
              surfelSetTransform[i] = t[i];
            if (updateInverse(surfelSetTransform))
              setId = false;
            else
              fprintf(stderr, "%s:%d: Surfel set transform has no full rank -- replacing it by identity.\n",
                      __FILE__, __LINE__);
          }

        if (setId)
          setIdentity32(surfelSetTransform);
      }

      // Determine number of markers.
      surfelSetNumMarkers = queryFieldNumElements(surfelSet,
                                                  SFLTAG_SURFEL_SET_MARKER_IDXVEC);
      if (surfelSetNumMarkers < 0)  surfelSetNumMarkers = 0;

      // Determine number of feature relations (experimental).
      surfelSetExpNumFeatureRels = queryFieldNumElements(surfelSet,
                                                         SFLTAG_SURFEL_SET_EXP_FEATURE_RELS);
      if (surfelSetExpNumFeatureRels < 0)
        surfelSetExpNumFeatureRels = 0;
      else
        surfelSetExpNumFeatureRels /= 2; // We count relations, not indices!

      // Determine number of SRF calibrations (experimental).
      surfelSetSRFNumCalibrations = queryFieldNumElements(surfelSet,
                                                          SFLTAG_SURFEL_SET_SRF_CALIBFLOATS);
      if (surfelSetSRFNumCalibrations < 0)
        surfelSetSRFNumCalibrations = 0;
      else
        surfelSetSRFNumCalibrations /= 12; // We count calibrations, not floats!

      // Set up read transformation
      {
        double  readTransform[32];
        updateInverse(appCoordSystem);
        //fprintf(stderr, "-------------- %s:%d\n", __FILE__, __LINE__);
        //dumpMat32(stderr, appCoordSystem);
        //dumpMat32(stderr, fileCoordSystem);
        mulMat32Inv32(readTransform, appCoordSystem, fileCoordSystem);
        //fprintf(stderr, "-------------- %s:%d\n", __FILE__, __LINE__);
        //dumpMat32(stderr, readTransform);
        if (readWorldFlag)
          mulMat32Inv32(readTransform, surfelSetTransform, readTransform);
        //fprintf(stderr, "-------------- %s:%d\n", __FILE__, __LINE__);
        //dumpMat32(stderr, readTransform);

        surfelBuffer.setReadTransform(readTransform);
      }

      // Initialize KillerSurfel
      surfelBuffer.setPropertyMask(surfelSetProperties, diskSize, numProps, propIdcs, propOffs);
      setSurfelSetPropertyHints();

      return true;
    }
  return false;
}

///////////////////////////////////////////////////////////////////////////////

char  *InStream::getSurfelSetIdentifier(void) const
{
  if (queryField(surfelSet, SFLTAG_SURFEL_SET_IDENTIFIER))
    return getFieldString(surfelSet, SFLTAG_SURFEL_SET_IDENTIFIER);
  else
    {
      char *res = (char *)malloc(1024);
      if (res)
        {
          getSurfelSetIdentifier(res, 1024);
          res = (char *)realloc(res, strlen(res)+1);
        }
      return res;
    }
}

bool   InStream::getSurfelSetIdentifier(char *dest, size_t buflen) const
{
  if (queryField(surfelSet, SFLTAG_SURFEL_SET_IDENTIFIER))
    return getFieldString(surfelSet, SFLTAG_SURFEL_SET_IDENTIFIER, dest, buflen);
  else
    {
      snprintf(dest, buflen, "surfelSet%02d", surfelSetIdx);
      return true;
    }
}

///////////////////////////////////////////////////////////////////////////////

bool  InStream::seekResolution(int idx)
{
  if (surfelSetIdx >=0 && idx >= 0 && idx < (int)surfelSetNumResolutions)
    {
      if ((resolution = gff_read_child_dir(gff, surfelSet, resolutionIdx=idx)))
        {
          // Read number of surfels.
          resolutionNumSurfels = getFieldUInt32(resolution,
                                                SFLTAG_SURFEL_RSL_NUM_SURFELS,
                                                true, 0);
          rwBufferHead = rwBufferTail = rwBuffer[0];

          gff_seek_field_data(gff, resolution, SFLTAG_SURFEL_RSL_DATA);

          return true;
        }
    }
  return false;
}

///////////////////////////////////////////////////////////////////////////////

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
