// ### CHECK-ME-BEFORE-RELEASE ###
//
// Title:   killersurfel.cpp
// Created: Sat Apr 27 18:32:45 2002
// Authors: Tim Weyrich <weyrich@inf.ethz.ch>
//
// Copyright (c) 2001--2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//
// ---------------------------------------------------------------
//
// $Log: killersurfel.cpp,v $
// Revision 1.18  2004/04/27 15:34:06  weyrich
// experimental: fixing propmask/layout inconsistencies caused by the Pointshop3D writer
//
// Revision 1.17  2004/04/27 14:51:02  weyrich
// experimental new policy: dumbly write tangent axes, normals, and radii in parallel
//
// Revision 1.16  2004/04/27 13:58:46  weyrich
// fixed cut-n-paste error for shininess dump; added experimental r+t2 support
//
// Revision 1.15  2004/03/23 12:52:12  weyrich
// added density to tangent axes conversion
//
// Revision 1.14  2003/11/19 13:57:55  weyrich
// source code formatting
//
// Revision 1.13  2003/11/19 13:04:02  weyrich
// added Pointshop3D headers
//
// Revision 1.12  2003/11/19 09:50:49  weyrich
// fix: g++ optimization bug
//
// Revision 1.11  2003/11/18 09:44:45  weyrich
// fix: shininess output; new: clamping of double->uintXX on output
//
// Revision 1.10  2003/11/17 15:05:52  weyrich
// new API allowing for more than 32 surfel properties
//
// Revision 1.9  2003/10/09 15:40:57  weyrich
// added properties 'camera_id' and 'camera_weight' for Stephan Wuermlin
//
// Revision 1.8  2003/10/09 14:58:42  wickem
// added isovalue + cell fields
//
// Revision 1.7  2003/03/20 14:52:52  weyrich
// smaller fixes of implicit coordinate system transformations
//
// Revision 1.6  2002/11/26 18:42:03  weyrich
// fix: mis-aligned float slurp
//
// Revision 1.5  2002/11/18 13:09:12  weyrich
// added markers and tangential vectors
//
// Revision 1.4  2002/06/13 14:24:20  weyrich
// implemented YUV conversions
//
// Revision 1.3  2002/05/23 13:59:56  weyrich
// First official release.
//
// Revision 1.2  2002/05/17 13:58:16  weyrich
// PointShop3D integration nearly finished
//
// Revision 1.1.1.1  2002/05/08 08:58:47  weyrich
// surfel file format
//

#include <pistring.h>
#include <piendian.h>
#include <pimath.h>
#include <stddef.h>             // for offsetof()
#include <assert.h>

#include "api.h"
#include "internal/sflmat.h"

using namespace sfl;
using namespace sflMat;

#ifdef _WIN32
/* disable "unary minus operator applied to unsigned type, result still unsigned" */
#  pragma warning(disable : 4146)
#endif

// KNOWN BUGS (FIXME!!):
//
//   Shininess isn't stored/read.
//
//   Problems might occur, if sampling density is stored and
//   tangentAxes are read.
//

// TODO: FIXME: surfelPropMask: Perhaps also enable radius, if sampling density is present.

#define EXPERIMENTAL_RADIUS_AND_TANGENTIAL

///////////////////////////////////////////////////////////////////////////////

const float  rgbToYuvMat[3][3] =
  {
    {  0.299f,  0.587f,  0.114f},
    { -0.147f, -0.289f,  0.437f},
    {  0.615f, -0.515f, -0.100f}
  };

const float  yuvToRgbMat[3][3] =
  {
    {  1.f,  0.f   ,  1.140f },
    {  1.f, -0.394f, -0.581f },
    {  1.f,  2.028f,  0.f    }
  };

const float  rgbToYiqMat[3][3] =
  {
    {  0.299f,  0.587f,  0.114f },
    {  0.596f, -0.274f, -0.322f },
    {  0.212f, -0.523f,  0.311f }
  };

const float  yiqToRgbMat[3][3] =
  {
    {  1.f,  0.956f,  0.621f },
    {  1.f, -0.272f, -0.647f },
    {  1.f, -1.105f,  1.702f }
  };

///////////////////////////////////////////////////////////////////////////////

KillerSurfel::KillerSurfel()
  : surfelPadding(1),
    propertyMask(0x0),
    propertyHints(0x0),
    numConversions(0)
{
  memset(propertyOffset, 0, ((int)MAX_PROPERTY_BITS) * sizeof(propertyOffset[0]));
  memset(propertySize, 0, ((int)MAX_PROPERTY_BITS) * sizeof(propertySize[0]));

  {
    int  i;
    for (i=0; i<(int)MAX_PROPERTY_BITS; i++)
      propertyPadding[i] = 1;
  }

  propertySize[(int)NORMAL_BIT] = 3*sizeof(float);
  propertySize[(int)POSITION_BIT] = 3*sizeof(float);
  propertySize[(int)RADIUS_BIT] = sizeof(float);
  propertySize[(int)WEIGHT_BIT] = sizeof(float);
  propertySize[(int)SAMPLING_DENSITY_BIT] = 3*sizeof(float);
  propertySize[(int)TEXTURE_UV_BIT] = 2*sizeof(float);
  propertySize[(int)AMBIENT_COLOR_BIT] = 0;
  propertySize[(int)DIFFUSE_COLOR_BIT] = 0;
  propertySize[(int)SPECULAR_COLOR_BIT] = 0;
  propertySize[(int)AMBIENT_COEFF_BIT] = sizeof(uint8);
  propertySize[(int)DIFFUSE_COEFF_BIT] = sizeof(uint8);
  propertySize[(int)SPECULAR_COEFF_BIT] = sizeof(uint8);
  propertySize[(int)SHININESS_BIT] = sizeof(uint8);
  propertySize[(int)TANGENT_AXES_BIT] = 6*sizeof(float);
  propertySize[(int)FLAGS_BIT] = sizeof(uint32);
  propertySize[(int)ISOVALUE_BIT] = sizeof(float);
  propertySize[(int)ORIGIN_CELL_BIT] = sizeof(uint32);
  propertySize[(int)CAMERA_DEPTH_BIT] = sizeof(float);
  propertySize[(int)CAMERA_ID_BIT] = sizeof(uint8);
  propertySize[(int)CAMERA_XY_BIT] = 2*sizeof(uint16);
  propertySize[(int)NORMAL_INFO_RAY_ONE_BIT] = 3*sizeof(float);
  propertySize[(int)NORMAL_INFO_RAY_TWO_BIT] = 3*sizeof(float);
  propertySize[(int)NORMAL_INFO_THETA_PHI_BIT] = 2*sizeof(float);

  propertyPadding[(int)NORMAL_BIT] = sizeof(float);
  propertyPadding[(int)POSITION_BIT] = sizeof(float);
  propertyPadding[(int)RADIUS_BIT] = sizeof(float);
  propertyPadding[(int)WEIGHT_BIT] = sizeof(float);
  propertyPadding[(int)SAMPLING_DENSITY_BIT] = sizeof(float);
  propertyPadding[(int)TEXTURE_UV_BIT] = sizeof(float);
  propertyPadding[(int)AMBIENT_COLOR_BIT] = 1;
  propertyPadding[(int)DIFFUSE_COLOR_BIT] = 1;
  propertyPadding[(int)SPECULAR_COLOR_BIT] = 1;
  propertyPadding[(int)AMBIENT_COEFF_BIT] = sizeof(uint8);
  propertyPadding[(int)DIFFUSE_COEFF_BIT] = sizeof(uint8);
  propertyPadding[(int)SPECULAR_COEFF_BIT] = sizeof(uint8);
  propertyPadding[(int)SHININESS_BIT] = sizeof(uint8);
  propertyPadding[(int)TANGENT_AXES_BIT] = sizeof(float);
  propertyPadding[(int)FLAGS_BIT] = sizeof(uint32);
  propertyPadding[(int)ISOVALUE_BIT] = sizeof(float);
  propertyPadding[(int)ORIGIN_CELL_BIT] = sizeof(uint32);
  propertyPadding[(int)CAMERA_DEPTH_BIT] = sizeof(float);
  propertyPadding[(int)CAMERA_ID_BIT] = sizeof(uint8);
  propertyPadding[(int)CAMERA_XY_BIT] = 2*sizeof(uint16);
  propertyPadding[(int)NORMAL_INFO_RAY_ONE_BIT] = 3*sizeof(float);
  propertyPadding[(int)NORMAL_INFO_RAY_TWO_BIT] = 3*sizeof(float);
  propertyPadding[(int)NORMAL_INFO_THETA_PHI_BIT] = 2*sizeof(float);

  spill = &KillerSurfel::spillGeneric;
  slurp = &KillerSurfel::slurpGeneric;
}

///////////////////////////////////////////////////////////////////////////////

void  KillerSurfel::updatePropertyLayout(int diskSizeOverride, int numProps, const uint16 *propIndcs, const uint32 *propOffs)
{
#ifndef EXPERIMENTAL_RADIUS_AND_TANGENTIAL
  if (propertyMask.isSet(TANGENT_AXES_BIT))
	{
	  propertyMask.clear(NORMAL_BIT);
	  propertyMask.clear(RADIUS_BIT);
	}
#endif
  
#ifndef EXPERIMENTAL_RADIUS_AND_TANGENTIAL
  if (propertyMask.isSet(TANGENT_AXES_BIT))
    {
      propertySize[(int)NORMAL_BIT] = 0;
      propertyPadding[(int)NORMAL_BIT] = 1;
    }
  else
#endif
  if (propertyMask.isSet(NORMAL_BIT))
    switch (getNormalEncoding())
      {
      case ICOSAHEDRAL_8:
        propertySize[(int)NORMAL_BIT] = 1;
        propertyPadding[(int)NORMAL_BIT] = 1;
        break;
      case CUBIC_16:
        propertySize[(int)NORMAL_BIT] = 2;
        propertyPadding[(int)NORMAL_BIT] = 2;
        break;
      case CUBIC_32:
        propertySize[(int)NORMAL_BIT] = 4;
        propertyPadding[(int)NORMAL_BIT] = 4;
        break;
      default:
        propertySize[(int)NORMAL_BIT] = 3*sizeof(float);
        propertyPadding[(int)NORMAL_BIT] = sizeof(float);
      }

  // Update colorSize.
  {
    switch (propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK)
      {
      case 0:  colorSize = 0; break;
      case SFLPROP_COLOR_MODEL_L:     colorSize = 1; break;
      case SFLPROP_COLOR_MODEL_LA:    colorSize = 2; break;
      case SFLPROP_COLOR_MODEL_RGB:   colorSize = 3; break;
      case SFLPROP_COLOR_MODEL_RGBA:  colorSize = 4; break;
      case SFLPROP_COLOR_MODEL_YUV:   colorSize = 3; break;
      case SFLPROP_COLOR_MODEL_YUVA:  colorSize = 4; break;
      default:
        fprintf(stderr, "%s:%d: WARNING: Unsupported color format.\n",
                __FILE__, __LINE__);
        colorSize = 0;
      }

    propertySize[(int)AMBIENT_COLOR_BIT] = colorSize;
    propertySize[(int)DIFFUSE_COLOR_BIT] = colorSize;
    propertySize[(int)SPECULAR_COLOR_BIT] = colorSize;
  }

  // Update propertyOffset[] and diskSize.
  {
    if (propIndcs)
      {
        int     i, j;
        for(i=0; i<(int)MAX_PROPERTY_BITS; i++)
          if (propertyMask.isSet(i) && propertySize[i])
            {
              for (j=0; j<numProps && i!=propIndcs[j]; j++);
              if (j < numProps)
                propertyOffset[i] = propOffs[j];
              else
                propertySize[i] = 0;
            }
        diskSize = diskSizeOverride;
      }
    else
      {
        int     p, i;
        size_t  offs = 0;
        for (p=8; p; p/=2)
          for (i=0; i<(int)MAX_PROPERTY_BITS; i++)
            if ((propertyMask.isSet(i) && propertySize[i]) &&
                ((int)propertyPadding[i]) == p)
              {
                offs = (offs + propertyPadding[i]-1) & -propertyPadding[i];
                propertyOffset[i] = offs;
                offs += propertySize[i];
              }
        offs = (offs + surfelPadding-1) & -surfelPadding;
        diskSize = diskSizeOverride >= 0 ? diskSizeOverride : offs;
        //fprintf(stderr, "DISK SPACE PER SURFEL: %d bytes\n", (int)diskSize);
      }
  }

  // Update layoutWriteXxxLookup[] tables.
  {
    int  c2cHead = 0;
    int  d2fHead = 0;
    int  d2u16Head = 0;
    int  d2u8Head = 0;

    //// FIRST SOME MACROS (will be substituted soon) /////////////////////////

#define  TBL_U8_U8(field, fieldIdx, idx)                                                        \
    do {                                                                                        \
      layoutWriteUInt8ToUInt8Lookup[c2cHead][0] = (int16)offsetof(KillerSurfel, field);         \
      layoutWriteUInt8ToUInt8Lookup[c2cHead][1] = (int16)propertyOffset[(fieldIdx)] + 1*idx;    \
      c2cHead++;                                                                                \
    } while(0)

#if PI_HAVE_NETWORK_ENDIAN

    //// BIG ENDIAN MACHINE ///////////////////////////////////////////////////

#  define  TBL_U16_U16(field, fieldIdx, idx)                                                            \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field);               \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 2*idx;     \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] + 1;    \
      c2cHead += 2;                                                                                     \
    } while(0)

#  define  TBL_U32_U8(field, fieldIdx, idx)                                                             \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field) + 3;           \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 1*idx;     \
      c2cHead++;                                                                                        \
    } while(0)

#  define  TBL_U32_U16(field, fieldIdx, idx)                                                            \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field) + 2;           \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 2*idx;     \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] + 1;    \
      c2cHead += 2;                                                                                     \
    } while(0)

#  define  TBL_U32_U32(field, fieldIdx, idx)                                                            \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field);               \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 4*idx;     \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+2][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 2;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+2][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] + 2;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+3][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 3;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+3][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] + 3;    \
      c2cHead += 4;                                                                                     \
    } while(0)

#else

    //// LITTLE ENDIAN MACHINE ////////////////////////////////////////////////

#  define  TBL_U16_U16(field, fieldIdx, idx)                                                            \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field);               \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 2*idx + 1; \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] - 1;    \
      c2cHead += 2;                                                                                     \
    } while(0)

#  define  TBL_U32_U8(field, fieldIdx, idx)   TBL_U8_U8(field, fieldIdx, idx)

#  define  TBL_U32_U16(field, fieldIdx, idx)  TBL_U16_U16(field, fieldIdx, idx)

#  define  TBL_U32_U32(field, fieldIdx, idx)                                                            \
    do {                                                                                                \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] = (int16)offsetof(KillerSurfel, field);               \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] = (int16)propertyOffset[(int)(fieldIdx)] + 4*idx + 3; \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+1][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] - 1;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+2][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 2;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+2][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] - 2;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+3][0] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][0] + 3;    \
      layoutWriteUInt8ToUInt8Lookup[c2cHead+3][1] = layoutWriteUInt8ToUInt8Lookup[c2cHead+0][1] - 3;    \
      c2cHead += 4;                                                                                     \
    } while(0)

#endif

    //// ALL MACHINES /////////////////////////////////////////////////////////

#define  TBL_F_F(field, fieldIdx, idx)  TBL_U32_U32(field, fieldIdx, idx)

#define  TBL_D_F(field, fieldIdx, idx)                                                                         \
    do {                                                                                                       \
      layoutWriteDoubleToFloatLookup[d2fHead][0] = (int16)offsetof(KillerSurfel, field);                       \
      layoutWriteDoubleToFloatLookup[d2fHead][1] = (int16)propertyOffset[(int)(fieldIdx)] + sizeof(float)*idx; \
      d2fHead++;                                                                                               \
    } while(0)

#define  TBL_D_U16(field, fieldIdx, idx)                                                                                   \
    do {                                                                                                           \
      layoutWriteDoubleToUInt16Lookup[d2u16Head][0] = (int16)offsetof(KillerSurfel, field);                        \
      layoutWriteDoubleToUInt16Lookup[d2u16Head][1] = (int16)propertyOffset[(int)(fieldIdx)] + sizeof(uint16)*idx; \
      d2u16Head++;                                                                                                 \
    } while(0)

#define  TBL_D_U8(field, fieldIdx, idx)                                                                            \
    do {                                                                                                           \
      layoutWriteDoubleToUInt8Lookup[d2u8Head][0] = (int16)offsetof(KillerSurfel, field);                          \
      layoutWriteDoubleToUInt8Lookup[d2u8Head][1] = (int16)propertyOffset[(int)(fieldIdx)] + sizeof(uint8)*idx; \
      d2u8Head++;                                                                                                  \
    } while(0)

#define  TBL_COL_U8_U8(field, fieldIdx)                                 \
    do {                                                                \
      switch (propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK)     \
        {                                                               \
        case SFLPROP_COLOR_MODEL_L:                                     \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          break;                                                        \
        case SFLPROP_COLOR_MODEL_LA:                                    \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          TBL_U8_U8(field[3], fieldIdx, 1);                             \
          break;                                                        \
        case SFLPROP_COLOR_MODEL_RGB:                                   \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          TBL_U8_U8(field[1], fieldIdx, 1);                             \
          TBL_U8_U8(field[2], fieldIdx, 2);                             \
          break;                                                        \
        case SFLPROP_COLOR_MODEL_RGBA:                                  \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          TBL_U8_U8(field[1], fieldIdx, 1);                             \
          TBL_U8_U8(field[2], fieldIdx, 2);                             \
          TBL_U8_U8(field[3], fieldIdx, 3);                             \
          break;                                                        \
        case SFLPROP_COLOR_MODEL_YUV:                                   \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          TBL_U8_U8(field[1], fieldIdx, 1);                             \
          TBL_U8_U8(field[2], fieldIdx, 2);                             \
          break;                                                        \
        case SFLPROP_COLOR_MODEL_YUVA:                                  \
          TBL_U8_U8(field[0], fieldIdx, 0);                             \
          TBL_U8_U8(field[1], fieldIdx, 1);                             \
          TBL_U8_U8(field[2], fieldIdx, 2);                             \
          TBL_U8_U8(field[3], fieldIdx, 3);                             \
          break;                                                        \
        default:                                                        \
          /* nop */ ;                                                   \
        }                                                               \
    } while(0)                                                          \

    ///////////////////////////////////////////////////////////////////////////

      // Code again:

      if (propertyMask.isSet(POSITION_BIT)) {
        TBL_D_F(position[0], POSITION_BIT, 0);
        TBL_D_F(position[1], POSITION_BIT, 1);
        TBL_D_F(position[2], POSITION_BIT, 2);
      }

      if (propertyMask.isSet(NORMAL_BIT))
        switch (getNormalEncoding())
          {
          case NONE:
            TBL_D_F(normal[0], NORMAL_BIT, 0);
            TBL_D_F(normal[1], NORMAL_BIT, 1);
            TBL_D_F(normal[2], NORMAL_BIT, 2);
            break;
          case ICOSAHEDRAL_8:
            TBL_U32_U8(normalCmpr.ico8, NORMAL_BIT, 0);
            break;
          case CUBIC_16:
            TBL_U32_U16(normalCmpr.cub16, NORMAL_BIT, 0);
            break;
          case CUBIC_32:
            TBL_U32_U32(normalCmpr.cub32, NORMAL_BIT, 0);
            break;
          }

      if (propertyMask.isSet(RADIUS_BIT))
        TBL_D_F(radius, RADIUS_BIT, 0);

      if (propertyMask.isSet(WEIGHT_BIT))
        TBL_D_F(weight, WEIGHT_BIT, 0);

      if (propertyMask.isSet(SAMPLING_DENSITY_BIT)) {
        TBL_D_F(samplingDensity[0], SAMPLING_DENSITY_BIT, 0);
        TBL_D_F(samplingDensity[1], SAMPLING_DENSITY_BIT, 1);
        TBL_D_F(samplingDensity[2], SAMPLING_DENSITY_BIT, 2);
      }

      if (propertyMask.isSet(TEXTURE_UV_BIT)) {
        TBL_D_F(textureUV[0], TEXTURE_UV_BIT, 0);
        TBL_D_F(textureUV[1], TEXTURE_UV_BIT, 1);
      }
	  
	  if (propertyMask.isSet(AMBIENT_COLOR_BIT))
		TBL_COL_U8_U8(colors[0], AMBIENT_COLOR_BIT);
	  
      if (propertyMask.isSet(AMBIENT_COEFF_BIT))
        TBL_U8_U8(coeffs[0], AMBIENT_COEFF_BIT, 0);
	  
	  if (propertyMask.isSet(DIFFUSE_COLOR_BIT))
		TBL_COL_U8_U8(colors[1], DIFFUSE_COLOR_BIT);
	  
      if (propertyMask.isSet(DIFFUSE_COEFF_BIT))
        TBL_U8_U8(coeffs[1], DIFFUSE_COEFF_BIT, 0);

      if (propertyMask.isSet(SPECULAR_COLOR_BIT))
        TBL_COL_U8_U8(colors[2], SPECULAR_COLOR_BIT);

      if (propertyMask.isSet(SPECULAR_COEFF_BIT))
        TBL_U8_U8(coeffs[2], SPECULAR_COEFF_BIT, 0);

      if (propertyMask.isSet(SHININESS_BIT))
        // At the moment, shininess is clamped to [0, 255] and dumped as uint8.
        TBL_D_U8(shininess, SHININESS_BIT, 0);
          
      if (propertyMask.isSet(TANGENT_AXES_BIT))
        {
          TBL_D_F(tangentAxes[0][0], TANGENT_AXES_BIT, 0);
          TBL_D_F(tangentAxes[0][1], TANGENT_AXES_BIT, 1);
          TBL_D_F(tangentAxes[0][2], TANGENT_AXES_BIT, 2);
          TBL_D_F(tangentAxes[1][0], TANGENT_AXES_BIT, 3);
          TBL_D_F(tangentAxes[1][1], TANGENT_AXES_BIT, 4);
          TBL_D_F(tangentAxes[1][2], TANGENT_AXES_BIT, 5);
        }

      if (propertyMask.isSet(FLAGS_BIT))
        {
          switch (propertySize[FLAGS_BIT])
            {
            case 1:
              TBL_U32_U8(flags, FLAGS_BIT, 0);
              break;
            case 2:
              TBL_U32_U16(flags, FLAGS_BIT, 0);
              break;
            case 4:
              TBL_U32_U32(flags, FLAGS_BIT, 0);
              break;
            }
        }

      if (propertyMask.isSet(ISOVALUE_BIT))
        TBL_D_F(isovalue, ISOVALUE_BIT, 0);

      if (propertyMask.isSet(ORIGIN_CELL_BIT))
        TBL_U32_U32(origin_cell, ORIGIN_CELL_BIT, 0);

      if (propertyMask.isSet(CAMERA_DEPTH_BIT))
        TBL_D_F(cameraDepth, CAMERA_DEPTH_BIT, 0);

      if (propertyMask.isSet(CAMERA_ID_BIT))
        TBL_U8_U8(cameraId, CAMERA_ID_BIT, 0);

      if (propertyMask.isSet(CAMERA_XY_BIT)) {
        TBL_D_U16(cameraXY[0], CAMERA_XY_BIT, 0);
        TBL_D_U16(cameraXY[1], CAMERA_XY_BIT, 1);
      }

      if (propertyMask.isSet(NORMAL_INFO_RAY_ONE_BIT)) {
        TBL_D_F(normalInfoRays[0][0], NORMAL_INFO_RAY_ONE_BIT, 0);
        TBL_D_F(normalInfoRays[0][1], NORMAL_INFO_RAY_ONE_BIT, 1);
        TBL_D_F(normalInfoRays[0][2], NORMAL_INFO_RAY_ONE_BIT, 2);
      }

      if (propertyMask.isSet(NORMAL_INFO_RAY_TWO_BIT)) {
        TBL_D_F(normalInfoRays[1][0], NORMAL_INFO_RAY_TWO_BIT, 0);
        TBL_D_F(normalInfoRays[1][1], NORMAL_INFO_RAY_TWO_BIT, 1);
        TBL_D_F(normalInfoRays[1][2], NORMAL_INFO_RAY_TWO_BIT, 2);
      }

      if (propertyMask.isSet(NORMAL_INFO_THETA_PHI_BIT)) {
        TBL_D_F(normalInfoThetaPhi[0], NORMAL_INFO_THETA_PHI_BIT, 0);
        TBL_D_F(normalInfoThetaPhi[1], NORMAL_INFO_THETA_PHI_BIT, 1);
      }

      layoutWriteUInt8ToUInt8Lookup[c2cHead][0] = -1;
      layoutWriteDoubleToFloatLookup[d2fHead][0] = -1;
      layoutWriteDoubleToUInt16Lookup[d2u16Head][0] = -1;
      layoutWriteDoubleToUInt8Lookup[d2u8Head][0] = -1;
  }
}

#define  PUSH_CONVERSION(c)  (conversions[numConversions++]=(c))

void  KillerSurfel::updateReadConversions(uint32 surfelSetFormatVersion)
{
  numConversions = 0;

  if (surfelSetFormatVersion == 0x00000000)
	{
	  // Version 0 did not allow for tangential axes and normals to be
	  // simultaneously written. Though, Pointshop3D tried to write
	  // both. Although all three bits were set in the propertyMask,
	  // only tangent axes were written.
	  //
	  if (propertyMask.isSet(TANGENT_AXES_BIT))
		{
		  propertyMask.clear(NORMAL_BIT);
		  propertyMask.clear(RADIUS_BIT);
		}
	}
  
  if ((propertyMask.isSet(NORMAL_BIT))
      && ((propertyHints.isSet(NORMAL_BIT))
          || (propertyHints.isSet(TANGENT_AXES_BIT))))
    {
      switch (getNormalEncoding())
        {
        case NONE:
          break;
        case ICOSAHEDRAL_8:
          PUSH_CONVERSION(&KillerSurfel::convIco8ToNormal);
          break;
        case CUBIC_16:
          PUSH_CONVERSION(&KillerSurfel::convCub16ToNormal);
          break;
        case CUBIC_32:
          PUSH_CONVERSION(&KillerSurfel::convCub32ToNormal);
          break;
        }
      if (!normalTransformIsId)
        {
          if (normalTransformIs3x3)
            PUSH_CONVERSION(&KillerSurfel::convNormalTransform3x3);
          else if (normalTransformIs3x4)
            PUSH_CONVERSION(&KillerSurfel::convNormalTransform3x4);
          else
            PUSH_CONVERSION(&KillerSurfel::convNormalTransform);
        }
    }

  if (propertyMask.isSet(TANGENT_AXES_BIT)
      && (((propertyHints.isSet(NORMAL_BIT) && !propertyMask.isSet(NORMAL_BIT))
		   || (propertyHints.isSet(RADIUS_BIT) && !propertyMask.isSet(RADIUS_BIT)))
          || (propertyHints.isSet(TANGENT_AXES_BIT))))
    {
      if (!positionTransformIsId)
        {
          // Note that the tangent transformations must be applied
          // *before* the position transformation.
          if (positionTransformIs3x3)
            if (positionTransformHasNegDet)
              PUSH_CONVERSION(&KillerSurfel::convTangentTransform3x3NegDet);
            else
              PUSH_CONVERSION(&KillerSurfel::convTangentTransform3x3);
          else if (positionTransformIs3x4)
            if (positionTransformHasNegDet)
              PUSH_CONVERSION(&KillerSurfel::convTangentTransform3x4NegDet);
            else
              PUSH_CONVERSION(&KillerSurfel::convTangentTransform3x4);
          else
            if (positionTransformHasNegDet)
              PUSH_CONVERSION(&KillerSurfel::convTangentTransformNegDet);
            else
              PUSH_CONVERSION(&KillerSurfel::convTangentTransform);
        }
    }

  if ((propertyMask.isSet(POSITION_BIT)) &&
      (propertyHints.isSet(POSITION_BIT)))
    {
      if (!positionTransformIsId)
        {
          if (positionTransformIs3x3)
            PUSH_CONVERSION(&KillerSurfel::convPositionTransform3x3);
          else if (positionTransformIs3x4)
            PUSH_CONVERSION(&KillerSurfel::convPositionTransform3x4);
          else
            PUSH_CONVERSION(&KillerSurfel::convPositionTransform);
        }
    }

  if ((propertyMask.isSet(RADIUS_BIT)) &&
      (propertyHints.isSet(SAMPLING_DENSITY_BIT)))
    if (propertyMask.isSet(NORMAL_BIT))
      PUSH_CONVERSION(&KillerSurfel::convRadiusToSamplingDensity);
    else
      PUSH_CONVERSION(&KillerSurfel::convRadiusToSamplingDensityWithoutNormal);

  if ((propertyMask.isSet(SAMPLING_DENSITY_BIT)) &&
      ((propertyHints.isSet(RADIUS_BIT))
       || propertyHints.isSet(TANGENT_AXES_BIT)))
    PUSH_CONVERSION(&KillerSurfel::convSamplingDensityToRadius);

  if (((propertyMask.isSet(NORMAL_BIT) &&
		propertyMask.isSet(RADIUS_BIT)) ||
	   propertyMask.isSet(SAMPLING_DENSITY_BIT)) &&
	  (!propertyMask.isSet(TANGENT_AXES_BIT)
	   && propertyHints.isSet(TANGENT_AXES_BIT)))
    PUSH_CONVERSION(&KillerSurfel::convRadiusNormalToTangentAxes);

  switch (propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK)
    {
    case SFLPROP_COLOR_MODEL_L:
    case SFLPROP_COLOR_MODEL_LA:
      switch ((propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK))
        {
        case SFLPROP_COLOR_MODEL_RGB:
        case SFLPROP_COLOR_MODEL_RGBA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelLToRGB);
          break;
        case SFLPROP_COLOR_MODEL_YUV:
        case SFLPROP_COLOR_MODEL_YUVA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelLToRGB);
          PUSH_CONVERSION(&KillerSurfel::convColorModelRGBToYUV);
          break;
        }
      break;

    case SFLPROP_COLOR_MODEL_RGB:
    case SFLPROP_COLOR_MODEL_RGBA:
      switch ((propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK))
        {
        case SFLPROP_COLOR_MODEL_L:
        case SFLPROP_COLOR_MODEL_LA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelRGBToL);
          break;
        case SFLPROP_COLOR_MODEL_YUV:
        case SFLPROP_COLOR_MODEL_YUVA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelRGBToYUV);
          break;
        }
      break;

    case SFLPROP_COLOR_MODEL_YUV:
    case SFLPROP_COLOR_MODEL_YUVA:
      switch ((propertyMask.getUInt32(0) & SFLPROP_COLOR_MODEL_MASK))
        {
        case SFLPROP_COLOR_MODEL_L:
        case SFLPROP_COLOR_MODEL_LA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelYUVToRGB);
          PUSH_CONVERSION(&KillerSurfel::convColorModelRGBToL);
          break;
        case SFLPROP_COLOR_MODEL_RGB:
        case SFLPROP_COLOR_MODEL_RGBA:
          PUSH_CONVERSION(&KillerSurfel::convColorModelYUVToRGB);
          break;
        }
      break;
    }
  // Set alphas to a default value:
  colors[0][3] = 255;
  colors[1][3] = 255;
  colors[2][3] = 255;

  // Define, what color should be "any"
  if (propertyMask.isSet(AMBIENT_COLOR_BIT))
    PUSH_CONVERSION(&KillerSurfel::convCopyColorAmbientToAny);
  else if (propertyMask.isSet(DIFFUSE_COLOR_BIT))
    PUSH_CONVERSION(&KillerSurfel::convCopyColorDiffuseToAny);
  else if (propertyMask.isSet(SPECULAR_COLOR_BIT))
    PUSH_CONVERSION(&KillerSurfel::convCopyColorSpecularToAny);
  else
    memset(&colors[0][0], 255, sizeof(colors)); // ### FIXME: Does this work?

  if (propertyMask.isSet(TANGENT_AXES_BIT))
    {
      if (propertyHints.isSet(NORMAL_BIT) && !propertyMask.isSet(NORMAL_BIT))
        PUSH_CONVERSION(&KillerSurfel::convTangentAxesToNormal);
      if (propertyHints.isSet(RADIUS_BIT) && !propertyMask.isSet(RADIUS_BIT))
        PUSH_CONVERSION(&KillerSurfel::convTangentAxesToRadius);
    }

  assert(numConversions < MAX_PROPERTY_BITS);

#if 0
  fprintf(stderr, "KillerSurfel: ==============================\n");
  dumpMat32(stderr, readTransform);
  dumpMat16(stderr, positionTransform);
  dumpMat16(stderr, normalTransform);
#endif
}

void  KillerSurfel::updateWriteConversions(void)
{
  numConversions = 0;

  if (propertyMask.isSet(NORMAL_BIT))
    switch ((normalenc_compression_enum)
            ((propertyMask.getUInt32(0) & SFLPROP_NORMAL_COMPR_MASK)
             >> SFLPROP_NORMAL_COMPR_SHIFT))
      {
      case NONE:
        break;
      case ICOSAHEDRAL_8:
        PUSH_CONVERSION(&KillerSurfel::convNormalToIco8);
        break;
      case CUBIC_16:
        PUSH_CONVERSION(&KillerSurfel::convNormalToCub16);
        break;
      case CUBIC_32:
        PUSH_CONVERSION(&KillerSurfel::convNormalToCub32);
        break;
      }

  // Set alphas to a default value:
  colors[0][3] = 255;
  colors[1][3] = 255;
  colors[2][3] = 255;

  assert(numConversions < MAX_PROPERTY_BITS);
}
#undef  PUSH_CONVERSION

///////////////////////////////////////////////////////////////////////////////

// FIXME (KNOWN BUG):
//
//   At the moment, only calling setPropertyMask() prepares the
//   KillerSurfel for beeing dumped.
//
//   Calling setPropertyMask() and then setPropertyHints() will
//   prepare it for beeing read from disk.
//
//   This state behaviour may lead to programming errors. However,
//   these methods shall only be called by InStream and OutStream!
//

KillerSurfel  &KillerSurfel::setPropertyMask(const PropMaskType &mask)
{
  return setPropertyMask(mask, -1, -1, NULL, NULL);
}

KillerSurfel  &KillerSurfel::setPropertyMask(const PropMaskType &mask,
                                             int diskSizeOverride,
                                             int numProps, const uint16 *propIndcs, const uint32 *propOffs)
{
  if (mask != propertyMask)
    {
      propertyMask = mask;
      updatePropertyLayout(diskSizeOverride, numProps, propIndcs, propOffs);
      updateWriteConversions(); // FIXME: Might interfer with updateReadConversions()!!
    }
  return *this;
}

KillerSurfel  &KillerSurfel::setPropertyHints(unsigned long hints,
											  uint32 surfelSetFormatVersion)
{
  return setPropertyHints(PropMaskType(hints), surfelSetFormatVersion);
}

KillerSurfel  &KillerSurfel::setPropertyHints(const PropMaskType &hints,
											  uint32 surfelSetFormatVersion)
{
  if (hints != propertyHints)
    {
      propertyHints = hints;
      updateReadConversions(surfelSetFormatVersion);  // FIXME: Might interfer with updateWriteConversions()!!
    }
  return *this;
}

KillerSurfel  &KillerSurfel::setReadTransform(double readTransform[32])
{
  cpyMat32(this->readTransform, readTransform);

  cpyMat16(positionTransform, readTransform);
  cpyTrp16(normalTransform, readTransform+16);

  positionTransformIsId = isIdentity(positionTransform);
  positionTransformIs3x3 = is3x3(positionTransform);
  positionTransformIs3x4 = is3x4(positionTransform);
  positionTransformHasNegDet = hasNegDet(positionTransform);

  normalTransformIsId = isIdentity(normalTransform);
  normalTransformIs3x3 = is3x3(normalTransform);
  normalTransformIs3x4 = is3x4(normalTransform);
  normalTransformHasNegDet = hasNegDet(normalTransform);

  return *this;
}

///////////////////////////////////////////////////////////////////////////////

void  KillerSurfel::convNormalToIco8(void)
{
  normalCmpr.ico8 = (uint8)normalenc_encode_normal_ico8(normal[0], normal[1], normal[2]);
}

void  KillerSurfel::convNormalToCub16(void)
{
  normalCmpr.cub16 = (uint16)normalenc_encode_normal_cub16(normal[0], normal[1], normal[2]);
}

void  KillerSurfel::convNormalToCub32(void)
{
  normalCmpr.cub32 = normalenc_encode_normal_cub32(normal[0], normal[1], normal[2]);
}

// ----------------------------------------------------------------------------

void  KillerSurfel::convIco8ToNormal(void)
{
  normalenc_decode_normal_ico8_3d(normalCmpr.ico8, &normal[0], &normal[1], &normal[2]);
}

void  KillerSurfel::convCub16ToNormal(void)
{
  normalenc_decode_normal_cub16_3d(normalCmpr.cub16, &normal[0], &normal[1], &normal[2]);
}

void  KillerSurfel::convCub32ToNormal(void)
{
  normalenc_decode_normal_cub32_3d(normalCmpr.cub32, &normal[0], &normal[1], &normal[2]);
}

#define  F_R  1.2f

void  KillerSurfel::convRadiusToSamplingDensity(void)
{
#if 1
  double  d = ((F_R*F_R)/(M_PI*radius*radius));
#else
  double  d = (1.0/(M_PI*(radius/F_R)*(radius/F_R)));
#endif
  samplingDensity[0] =  d * normal[0];
  samplingDensity[1] =  d * normal[1];
  samplingDensity[2] =  d * normal[2];
}

void  KillerSurfel::convRadiusToSamplingDensityWithoutNormal(void)
{
#if 1
  double  d = ((F_R*F_R)/(M_PI*radius*radius));
#else
  double  d = (1.0/(M_PI*(radius/F_R)*(radius/F_R)));
#endif
  samplingDensity[0] = d;
  samplingDensity[1] = 0.0;
  samplingDensity[2] = 0.0;
}

void  KillerSurfel::convSamplingDensityToRadius(void)
{
  radius = F_R / sqrt(M_PI * sqrt(samplingDensity[0]*samplingDensity[0] +
                                  samplingDensity[1]*samplingDensity[1] +
                                  samplingDensity[2]*samplingDensity[2]));
}

#define  R_WEIGHT  (30*65536/100)
#define  G_WEIGHT  (59*65536/100)
#define  B_WEIGHT  (11*65536/100)
#define  FIXP_SHIFT  16
void  KillerSurfel::convColorModelRGBToL(void)
{
  colors[0][0] = ((R_WEIGHT * colors[0][0]) +
                  (G_WEIGHT * colors[0][1]) +
                  (B_WEIGHT * colors[0][2])) >> FIXP_SHIFT;
  colors[1][0] = ((R_WEIGHT * colors[1][0]) +
                  (G_WEIGHT * colors[1][1]) +
                  (B_WEIGHT * colors[1][2])) >> FIXP_SHIFT;
  colors[2][0] = ((R_WEIGHT * colors[2][0]) +
                  (G_WEIGHT * colors[2][1]) +
                  (B_WEIGHT * colors[2][2])) >> FIXP_SHIFT;
}
#undef  R_WEIGHT
#undef  G_WEIGHT
#undef  B_WEIGHT
#undef  FIXP_SHIFT

void  KillerSurfel::convColorModelLToRGB(void)
{
  colors[0][1] = colors[0][0];
  colors[0][2] = colors[0][0];
  colors[1][1] = colors[1][0];
  colors[1][2] = colors[1][0];
  colors[2][1] = colors[2][0];
  colors[2][2] = colors[2][0];
}

// ----------------------------------------------------------------------------

static void  rgbToYuv(int r, int g, int b, int &y, int &u, int &v)
{
  y = (int)rint(rgbToYuvMat[0][0] * r + rgbToYuvMat[0][1] * g + rgbToYuvMat[0][2] * b);
  u = (int)rint(rgbToYuvMat[1][0] * r + rgbToYuvMat[1][1] * g + rgbToYuvMat[1][2] * b);
  v = (int)rint(rgbToYuvMat[2][0] * r + rgbToYuvMat[2][1] * g + rgbToYuvMat[2][2] * b);
}

static void  rgbToYuv(int r, int g, int b, unsigned char &y, unsigned char &u, unsigned char &v)
{
  int  Y, U, V;
  rgbToYuv(r, g, b, Y, U, V);
  y = (Y < 0) ? 0 : ((Y > 255) ? 255 : 0);
  u = (U < 0) ? 0 : ((U > 255) ? 255 : 0);
  v = (V < 0) ? 0 : ((V > 255) ? 255 : 0);
}

static void  yuvToRgb(int y, int u, int v, int &r, int &g, int &b)
{
  r = (int)rint(yuvToRgbMat[0][0] * y + yuvToRgbMat[0][1] * u + yuvToRgbMat[0][2] * v);
  g = (int)rint(yuvToRgbMat[1][0] * y + yuvToRgbMat[1][1] * u + yuvToRgbMat[1][2] * v);
  b = (int)rint(yuvToRgbMat[2][0] * y + yuvToRgbMat[2][1] * u + yuvToRgbMat[2][2] * v);
}

static void  yuvToRgb(int y, int u, int v, unsigned char &r, unsigned char &g, unsigned char &b)
{
  int  R, G, B;
  yuvToRgb(y, u, v, R, G, B);
  r = (R < 0) ? 0 : ((R > 255) ? 255 : 0);
  g = (G < 0) ? 0 : ((G > 255) ? 255 : 0);
  b = (B < 0) ? 0 : ((B > 255) ? 255 : 0);
}

// ----------------------------------------------------------------------

static void  rgbToYiq(int r, int g, int b, int &y, int &i, int &q)
{
  y = (int)rint(rgbToYiqMat[0][0] * r + rgbToYiqMat[0][1] * g + rgbToYiqMat[0][2] * b);
  i = (int)rint(rgbToYiqMat[1][0] * r + rgbToYiqMat[1][1] * g + rgbToYiqMat[1][2] * b);
  q = (int)rint(rgbToYiqMat[2][0] * r + rgbToYiqMat[2][1] * g + rgbToYiqMat[2][2] * b);
}

static void  rgbToYiq(int r, int g, int b, unsigned char &y, unsigned char &i, unsigned char &q)
{
  int  Y, I, Q;
  rgbToYiq(r, g, b, Y, I, Q);
  y = (Y < 0) ? 0 : ((Y > 255) ? 255 : 0);
  i = (I < 0) ? 0 : ((I > 255) ? 255 : 0);
  q = (Q < 0) ? 0 : ((Q > 255) ? 255 : 0);
}

static void  yiqToRgb(int y, int i, int q, int &r, int &g, int &b)
{
  r = (int)rint(yiqToRgbMat[0][0] * y + yiqToRgbMat[0][1] * i + yiqToRgbMat[0][2] * q);
  g = (int)rint(yiqToRgbMat[1][0] * y + yiqToRgbMat[1][1] * i + yiqToRgbMat[1][2] * q);
  b = (int)rint(yiqToRgbMat[2][0] * y + yiqToRgbMat[2][1] * i + yiqToRgbMat[2][2] * q);
}

static void  yiqToRgb(int y, int i, int q, unsigned char &r, unsigned char &g, unsigned char &b)
{
  int  R, G, B;
  yiqToRgb(y, i, q, R, G, B);
  r = (R < 0) ? 0 : ((R > 255) ? 255 : 0);
  g = (G < 0) ? 0 : ((G > 255) ? 255 : 0);
  b = (B < 0) ? 0 : ((B > 255) ? 255 : 0);
}

// ----------------------------------------------------------------------------

void  KillerSurfel::convColorModelRGBToYUV(void)
{
  rgbToYuv(colors[0][0], colors[0][1], colors[0][2],
           colors[0][0], colors[0][1], colors[0][2]);
  rgbToYuv(colors[1][0], colors[1][1], colors[1][2],
           colors[1][0], colors[1][1], colors[1][2]);
  rgbToYuv(colors[2][0], colors[2][1], colors[2][2],
           colors[2][0], colors[2][1], colors[2][2]);
}

void  KillerSurfel::convColorModelYUVToRGB(void)
{
  yuvToRgb(colors[0][0], colors[0][1], colors[0][2],
           colors[0][0], colors[0][1], colors[0][2]);
  yuvToRgb(colors[1][0], colors[1][1], colors[1][2],
           colors[1][0], colors[1][1], colors[1][2]);
  yuvToRgb(colors[2][0], colors[2][1], colors[2][2],
           colors[2][0], colors[2][1], colors[2][2]);
}

void  KillerSurfel::convColorModelRGBToYIQ(void)
{
  rgbToYiq(colors[0][0], colors[0][1], colors[0][2],
           colors[0][0], colors[0][1], colors[0][2]);
  rgbToYiq(colors[1][0], colors[1][1], colors[1][2],
           colors[1][0], colors[1][1], colors[1][2]);
  rgbToYiq(colors[2][0], colors[2][1], colors[2][2],
           colors[2][0], colors[2][1], colors[2][2]);
}

void  KillerSurfel::convColorModelYIQToRGB(void)
{
  yiqToRgb(colors[0][0], colors[0][1], colors[0][2],
           colors[0][0], colors[0][1], colors[0][2]);
  yiqToRgb(colors[1][0], colors[1][1], colors[1][2],
           colors[1][0], colors[1][1], colors[1][2]);
  yiqToRgb(colors[2][0], colors[2][1], colors[2][2],
           colors[2][0], colors[2][1], colors[2][2]);
}

void  KillerSurfel::convColorModelYUVToYIQ(void)
{
  convColorModelYUVToRGB();
  convColorModelRGBToYIQ();
}

void  KillerSurfel::convColorModelYIQToYUV(void)
{
  convColorModelYIQToRGB();
  convColorModelRGBToYUV();
}

void  KillerSurfel::convCopyColorAmbientToAny(void)
{
  *(uint32 *)&colors[3][0] = *(uint32 *)&colors[0][0];
}

void  KillerSurfel::convCopyColorDiffuseToAny(void)
{
  *(uint32 *)&colors[3][0] = *(uint32 *)&colors[1][0];
}

void  KillerSurfel::convCopyColorSpecularToAny(void)
{
  *(uint32 *)&colors[3][0] = *(uint32 *)&colors[2][0];
}

// ----------------------------------------------------------------------------

static void  cProd(const double a[3], const double b[3], double c[3])
{
  c[0] = a[1]*b[2] - a[2]*b[1];
  c[1] = a[2]*b[0] - a[0]*b[2];
  c[2] = a[0]*b[1] - a[1]*b[0];
}

static void  getTwoOrthogonals(const double n[3], double u[3], double v[3])
{
  if (fabs(n[0]) < fabs(n[1]))
    if (fabs(n[0]) < fabs(n[2]))
      { // n[0] is minimum
        u[0] =  0;
        u[1] = -n[2];
        u[2] =  n[1];
      }
    else
      { // n[2] is mimimum
        u[0] = -n[1];
        u[1] =  n[0];
        u[2] =  0;
      }
  else
    if (fabs(n[1]) < fabs(n[2]))
      { // n[1] is minimum
        u[0] =  n[2];
        u[1] =  0;
        u[2] = -n[0];
      }
    else
      { // n[2] is mimimum
        u[0] = -n[1];
        u[1] =  n[0];
        u[2] =  0;
      }
  cProd(n, u, v);
}

void  KillerSurfel::convRadiusNormalToTangentAxes(void)
{
  double  a;
  
  getTwoOrthogonals(normal, tangentAxes[0], tangentAxes[1]);
  
  a = sqrt(tangentAxes[0][0]*tangentAxes[0][0] +
           tangentAxes[0][1]*tangentAxes[0][1] +
           tangentAxes[0][2]*tangentAxes[0][2]);
  tangentAxes[0][0] *= radius/a;
  tangentAxes[0][1] *= radius/a;
  tangentAxes[0][2] *= radius/a;
  
  a = sqrt(tangentAxes[1][0]*tangentAxes[1][0] +
           tangentAxes[1][1]*tangentAxes[1][1] +
           tangentAxes[1][2]*tangentAxes[1][2]);
  tangentAxes[1][0] *= radius/a;
  tangentAxes[1][1] *= radius/a;
  tangentAxes[1][2] *= radius/a;
}

void  KillerSurfel::convTangentAxesToRadius(void)
{
  double  a, b;
  a = sqrt(tangentAxes[0][0]*tangentAxes[0][0] +
           tangentAxes[0][1]*tangentAxes[0][1] +
           tangentAxes[0][2]*tangentAxes[0][2]);
  b = sqrt(tangentAxes[1][0]*tangentAxes[1][0] +
           tangentAxes[1][1]*tangentAxes[1][1] +
           tangentAxes[1][2]*tangentAxes[1][2]);
  radius = a > b ? a : b;
}

void  KillerSurfel::convTangentAxesToNormal(void)
{
  double  a;
  cProd(tangentAxes[0], tangentAxes[1], normal);
  a = sqrt(normal[0]*normal[0] + normal[1]*normal[1] + normal[2]*normal[2]);
  normal[0] /= a;
  normal[1] /= a;
  normal[2] /= a;
}

// ============================================================================

#define  APPLY3X3(T,x)                          \
do                                              \
{                                               \
  double  a, b, c;                              \
  a = ((x)[0] * (T)[0] +                        \
       (x)[1] * (T)[1] +                        \
       (x)[2] * (T)[2]);                        \
  b = ((x)[0] * (T)[4] +                        \
       (x)[1] * (T)[5] +                        \
       (x)[2] * (T)[6]);                        \
  c = ((x)[0] * (T)[8] +                        \
       (x)[1] * (T)[9] +                        \
       (x)[2] * (T)[10]);                       \
  (x)[0] = a;                                   \
  (x)[1] = b;                                   \
  (x)[2] = c;                                   \
} while(0)

#define  APPLY3X4(T,x)                          \
do                                              \
{                                               \
  double  a, b, c;                              \
  a = ((x)[0] * (T)[0] +                        \
       (x)[1] * (T)[1] +                        \
       (x)[2] * (T)[2] + (T)[3]);               \
  b = ((x)[0] * (T)[4] +                        \
       (x)[1] * (T)[5] +                        \
       (x)[2] * (T)[6] + (T)[7]);               \
  c = ((x)[0] * (T)[8] +                        \
       (x)[1] * (T)[9] +                        \
       (x)[2] * (T)[10] + (T)[11]);             \
  (x)[0] = a;                                   \
  (x)[1] = b;                                   \
  (x)[2] = c;                                   \
} while(0)

#define  APPLY(T,x)                                     \
do                                                      \
{                                                       \
  double  a, b, c;                                      \
  a = ((x)[0] * (T)[0] +                                \
       (x)[1] * (T)[1] +                                \
       (x)[2] * (T)[2] + (T)[3]);                       \
  b = ((x)[0] * (T)[4] +                                \
       (x)[1] * (T)[5] +                                \
       (x)[2] * (T)[6] + (T)[7]);                       \
  c = ((x)[0] * (T)[8] +                                \
       (x)[1] * (T)[9] +                                \
       (x)[2] * (T)[10] + (T)[11]);                     \
  double f = 1.0 / ((x)[0] * (T)[12] +                  \
                    (x)[1] * (T)[13] +                  \
                    (x)[2] * (T)[14] + (T)[15]);        \
  (x)[0] = a*f;                                         \
  (x)[1] = b*f;                                         \
  (x)[2] = c*f;                                         \
} while(0)

void  KillerSurfel::convPositionTransform3x3(void)
{
  APPLY3X3(positionTransform, position);
}

void  KillerSurfel::convPositionTransform3x4(void)
{
  APPLY3X4(positionTransform, position);
}

void  KillerSurfel::convPositionTransform(void)
{
  APPLY(positionTransform, position);
}

void  KillerSurfel::convNormalTransform3x3(void)
{
  APPLY3X3(normalTransform, normal);
  double  f = 1.0/sqrt(normal[0]*normal[0] + normal[1]*normal[1] + normal[2]*normal[2]);
  normal[0] *= f;
  normal[1] *= f;
  normal[2] *= f;
}

void  KillerSurfel::convNormalTransform3x4(void)
{
  APPLY3X4(normalTransform, normal);
  double  f = 1.0/sqrt(normal[0]*normal[0] + normal[1]*normal[1] + normal[2]*normal[2]);
  normal[0] *= f;
  normal[1] *= f;
  normal[2] *= f;
}

void  KillerSurfel::convNormalTransform(void)
{
  APPLY(normalTransform, normal);
  double  f = 1.0/sqrt(normal[0]*normal[0] + normal[1]*normal[1] + normal[2]*normal[2]);
  normal[0] *= f;
  normal[1] *= f;
  normal[2] *= f;
}

void  KillerSurfel::convTangentTransform3x3(void)
{
  APPLY3X3(positionTransform, tangentAxes[0]);
  APPLY3X3(positionTransform, tangentAxes[1]);
}

void  KillerSurfel::convTangentTransform3x4(void)
{
  APPLY3X3(positionTransform, tangentAxes[0]);
  APPLY3X3(positionTransform, tangentAxes[1]);
}

/*! ATTENTION: This conversion must be applied \e before
 *  convPositionTransform()!!
 */
void  KillerSurfel::convTangentTransform(void)
{
  double  temp[3] = {position[0], position[1], position[2]};
  APPLY(positionTransform, temp);
  
  int  i;
  for(i=0; i<2; i++)
    {
      tangentAxes[i][0] += position[0];
      tangentAxes[i][1] += position[1];
      tangentAxes[i][2] += position[2];
      APPLY(positionTransform, tangentAxes[i]);
      tangentAxes[i][0] -= temp[0];
      tangentAxes[i][1] -= temp[1];
      tangentAxes[i][2] -= temp[2];
    }
}

void  KillerSurfel::convTangentTransform3x3NegDet(void)
{
  APPLY3X3(positionTransform, tangentAxes[0]);
  APPLY3X3(positionTransform, tangentAxes[1]);
  tangentAxes[1][0] = -tangentAxes[1][0];
  tangentAxes[1][1] = -tangentAxes[1][1];
  tangentAxes[1][2] = -tangentAxes[1][2];
}

void  KillerSurfel::convTangentTransform3x4NegDet(void)
{
  APPLY3X3(positionTransform, tangentAxes[0]);
  APPLY3X3(positionTransform, tangentAxes[1]);
  tangentAxes[1][0] = -tangentAxes[1][0];
  tangentAxes[1][1] = -tangentAxes[1][1];
  tangentAxes[1][2] = -tangentAxes[1][2];
}

/*! ATTENTION: This conversion must be applied \e before
 *  convPositionTransform()!!
 */
void  KillerSurfel::convTangentTransformNegDet(void)
{
  double  temp[3] = {position[0], position[1], position[2]};
  APPLY(positionTransform, temp);
  
  int  i;
  for(i=0; i<2; i++)
    {
      tangentAxes[i][0] += position[0];
      tangentAxes[i][1] += position[1];
      tangentAxes[i][2] += position[2];
      APPLY(positionTransform, tangentAxes[i]);
      tangentAxes[i][0] -= temp[0];
      tangentAxes[i][1] -= temp[1];
      tangentAxes[i][2] -= temp[2];
    }
  tangentAxes[1][0] = -tangentAxes[1][0];
  tangentAxes[1][1] = -tangentAxes[1][1];
  tangentAxes[1][2] = -tangentAxes[1][2];
}

///////////////////////////////////////////////////////////////////////////////

uint32  KillerSurfel::spillGeneric(uint8 *dest) const
{
  int  i;
  for (i=0; layoutWriteDoubleToFloatLookup[i][0]>0; i++)
    {
      // ### FIXME: This won't work... [### TODO: Can't remember why... :-( ]
      float *f = (float *)(((char *)dest) + layoutWriteDoubleToFloatLookup[i][1]);
      *f = (float)*(double *)(((const char *)this) + layoutWriteDoubleToFloatLookup[i][0]);
      pi_float_to_net_ip(f);
    }
  for (i=0; layoutWriteDoubleToUInt16Lookup[i][0]>0; i++)
    {
      uint16 *u = (uint16 *)(((char *)dest) + layoutWriteDoubleToUInt16Lookup[i][1]);
      *u = clampToUInt16(*(double *)(((const char *)this) + layoutWriteDoubleToUInt16Lookup[i][0]) + 0.5);
      pi_uint16_to_net_ip(u);
    }
  for (i=0; layoutWriteDoubleToUInt8Lookup[i][0]>0; i++)
    {
      uint8 *u = (uint8 *)(((char *)dest) + layoutWriteDoubleToUInt8Lookup[i][1]);
      *u = clampToUInt8(*(double *)(((const char *)this) + layoutWriteDoubleToUInt8Lookup[i][0]) + 0.5);
    }
  for (i=0; layoutWriteUInt8ToUInt8Lookup[i][0]>0; i++)
    ((char *)dest)[layoutWriteUInt8ToUInt8Lookup[i][1]] =
      ((const char *)this)[layoutWriteUInt8ToUInt8Lookup[i][0]];
  return diskSize;
}

uint32  KillerSurfel::slurpGeneric(uint8 *src) const
{
  int  i;
  for (i=0; layoutWriteDoubleToFloatLookup[i][0]>0; i++)
    {
#if 1
      float  f;
      memcpy(&f, ((char *)src) + layoutWriteDoubleToFloatLookup[i][1], sizeof(float));
#else
      float f = *(float *)(((char *)src) + layoutWriteDoubleToFloatLookup[i][1]);
#endif
      pi_float_to_net_ip(&f);
      *(double *)(((const char *)this) + layoutWriteDoubleToFloatLookup[i][0]) = f;
    }
  for (i=0; layoutWriteDoubleToUInt16Lookup[i][0]>0; i++)
    {
      uint16  u;
      memcpy(&u, ((char *)src) + layoutWriteDoubleToUInt16Lookup[i][1], sizeof(uint16));
      pi_uint16_to_net_ip(&u);
      *(double *)(((const char *)this) + layoutWriteDoubleToUInt16Lookup[i][0]) = u;
    }
  for (i=0; layoutWriteDoubleToUInt8Lookup[i][0]>0; i++)
    {
      uint8  u;
      memcpy(&u, ((char *)src) + layoutWriteDoubleToUInt8Lookup[i][1], sizeof(uint8));
      *(double *)(((const char *)this) + layoutWriteDoubleToUInt8Lookup[i][0]) = u;
    }
  for (i=0; layoutWriteUInt8ToUInt8Lookup[i][0]>0; i++)
    ((char *)this)[layoutWriteUInt8ToUInt8Lookup[i][0]] =
      ((const char *)src)[layoutWriteUInt8ToUInt8Lookup[i][1]];
  return diskSize;
}

///////////////////////////////////////////////////////////////////////////////

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
