// ### CHECK-ME-BEFORE-RELEASE ###
//
// Title:   stream.cpp
// Created: Thu Apr 18 14:29:28 2002
// Authors: Tim Weyrich <weyrich@inf.ethz.ch>
//
// Copyright (c) 2001--2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
//
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//
// ---------------------------------------------------------------
//
// $Log: stream.cpp,v $
// Revision 1.4  2003/11/19 13:04:02  weyrich
// added Pointshop3D headers
//
// Revision 1.3  2002/05/23 13:59:56  weyrich
// First official release.
//
// Revision 1.2  2002/05/17 13:58:16  weyrich
// PointShop3D integration nearly finished
//
// Revision 1.1.1.1  2002/05/08 08:58:47  weyrich
// surfel file format
//

#include <pistring.h>
#include <pimath.h>
#include <assert.h>

#include "api.h"
#include "internal/sflmat.h"

using namespace sfl;
using namespace sflMat;

#define  BUFFER_HALF_SIZE  8192

///////////////////////////////////////////////////////////////////////////////

void  sfl::memFree(void *mem)
{
  free(mem);
}

///////////////////////////////////////////////////////////////////////////////

Stream::Stream()
  : gff(NULL),
    rwBufferSize(BUFFER_HALF_SIZE)
{
  rwBuffer[0] = (uint8*)malloc(2*rwBufferSize*sizeof(uint8));
  assert(rwBuffer[0]);
  rwBuffer[1] = rwBuffer[0] + rwBufferSize;
  rwBufferHead = rwBuffer[0];
  rwBufferTail = rwBuffer[0];
}

Stream::~Stream()
{
  if (gff)  gff_close(gff);
}

///////////////////////////////////////////////////////////////////////////////

void  Stream::apiSwapMat4x4(float mat[16]) const
{
  if (!apiRowMajorMatrices)
    trpMat16(mat);
}

void  Stream::apiSwapMat4x4(double mat[16]) const
{
  if (!apiRowMajorMatrices)
    trpMat16(mat);
}

void  Stream::apiSwapMat4x4ToTemp4x4(float temp[16], const float mat[16]) const
{
  if (apiRowMajorMatrices)
    cpyMat16(temp, mat);
  else
    cpyTrp16(temp, mat);
}

void  Stream::apiSwapMat4x4ToTemp4x4(double temp[16], const double mat[16]) const
{
  if (apiRowMajorMatrices)
    cpyMat16(temp, mat);
  else
    cpyTrp16(temp, mat);
}


///////////////////////////////////////////////////////////////////////////////

// Although, a decomposition of the transformation to
//
//    T * R * S
//
// would be more common, PointShop seems to use
//
//    T*S * R
//
// So, we are using the latter...
//

static void  getScalation(const float *mat, float &sx, float &sy, float &sz)
{
  sx = (float)sqrt(mat[0]*mat[0] + mat[1]*mat[1] + mat[ 2]*mat[ 2]);
  sy = (float)sqrt(mat[4]*mat[4] + mat[5]*mat[5] + mat[ 6]*mat[ 6]);
  sz = (float)sqrt(mat[8]*mat[8] + mat[9]*mat[9] + mat[10]*mat[10]);
}

#if 0
static void  getScalation(const double *mat, float &sx, float &sy, float &sz)
{
  sx = (float)sqrt(mat[0]*mat[0] + mat[1]*mat[1] + mat[ 2]*mat[ 2]);
  sy = (float)sqrt(mat[4]*mat[4] + mat[5]*mat[5] + mat[ 6]*mat[ 6]);
  sz = (float)sqrt(mat[8]*mat[8] + mat[9]*mat[9] + mat[10]*mat[10]);
}
#endif

static void  getScalation(const double *mat, double &sx, double &sy, double &sz)
{
  sx = sqrt(mat[0]*mat[0] + mat[1]*mat[1] + mat[ 2]*mat[ 2]);
  sy = sqrt(mat[4]*mat[4] + mat[5]*mat[5] + mat[ 6]*mat[ 6]);
  sz = sqrt(mat[8]*mat[8] + mat[9]*mat[9] + mat[10]*mat[10]);
}

bool  Stream::apiGetTransform4x4(const double *src, float *mat, bool internal) const
{
  for(int i=0; i<16; i++)
    mat[i] = (float)src[i];
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetTransform4x4(const double *src, double *mat, bool internal) const
{
  memcpy(mat, &src[0], 16*sizeof(double));
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetInvTransform4x4(const double *src, float *mat, bool internal) const
{
  for(int i=0; i<16; i++)
    mat[i] = (float)src[i+16];
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetInvTransform4x4(const double *src, double *mat, bool internal) const
{
  memcpy(mat, &src[16], 16*sizeof(double));
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetRotation4x4(const double *src, float *mat, bool internal) const
{
  float  sx, sy, sz;
  apiGetScalationRotation4x4(src, mat, true);
  getScalation(mat, sx, sy, sz);
  mat[0] /= sx;  mat[1] /= sy;  mat[ 2] /= sz;
  mat[4] /= sx;  mat[5] /= sy;  mat[ 6] /= sz;
  mat[8] /= sx;  mat[9] /= sy;  mat[10] /= sz;
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetRotation4x4(const double *src, double *mat, bool internal) const
{
  double  sx, sy, sz;
  apiGetScalationRotation4x4(src, mat, true);
  getScalation(mat, sx, sy, sz);
  mat[0] /= sx;  mat[1] /= sy;  mat[ 2] /= sz;
  mat[4] /= sx;  mat[5] /= sy;  mat[ 6] /= sz;
  mat[8] /= sx;  mat[9] /= sy;  mat[10] /= sz;
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetScalationRotation4x4(const double *src, float *mat, bool internal) const
{
  int  i, j;
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      mat[i*4 + j] = (float)src[i*4 + j];
  mat[ 3] = 0.0f;
  mat[ 7] = 0.0f;
  mat[11] = 0.0f;

  mat[12] = 0.0f;
  mat[13] = 0.0f;
  mat[14] = 0.0f;
  mat[15] = 1.0f;

  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetScalationRotation4x4(const double *src, double *mat, bool internal) const
{
  int  i, j;
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      mat[i*4 + j] = src[i*4 + j];
  mat[ 3] = 0.0;
  mat[ 7] = 0.0;
  mat[11] = 0.0;

  mat[12] = 0.0;
  mat[13] = 0.0;
  mat[14] = 0.0;
  mat[15] = 1.0;

  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetTranslation4x4(const double *src, float *mat, bool internal) const
{
  setIdentity16(mat);
  mat[ 3] = (float)src[ 3];
  mat[ 7] = (float)src[ 7];
  mat[11] = (float)src[11];
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetTranslation4x4(const double *src, double *mat, bool internal) const
{
  setIdentity16(mat);
  mat[ 3] = src[ 3];
  mat[ 7] = src[ 7];
  mat[11] = src[11];
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetScalationTranslation4x4(const double *src, float *mat, bool internal) const
{
  double  rot[32];
  double  res[32];
  apiGetRotation4x4(src, rot, true);
  updateInverse(rot);
  mulMat32Inv32(res, src, rot);
  cpyMat16(mat, res);
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

bool  Stream::apiGetScalationTranslation4x4(const double *src, double *mat, bool internal) const
{
  double  rot[32];
  double  res[32];
  apiGetRotation4x4(src, rot, true);
  updateInverse(rot);
  mulMat32Inv32(res, src, rot);
  cpyMat16(mat, res);
  if (!internal)  apiSwapMat4x4(mat);
  return true;
}

///////////////////////////////////////////////////////////////////////////////

uint32  Stream::fillRwBuffer(void)
{
  if (rwBufferHead > rwBufferTail) // Due to an error on Win32
    memcpy(rwBuffer[0], rwBufferTail, rwBufferHead-rwBufferTail);
  else
    assert(rwBufferHead == rwBufferTail); // FIXME: Assertion appears on EOF!!
  rwBufferHead = rwBuffer[0] + (rwBufferHead - rwBufferTail);
  rwBufferTail = rwBuffer[0];
  assert(rwBufferHead <= rwBuffer[1]);
  ssize_t n = gff_read(gff, rwBufferHead, rwBufferSize);
  rwBufferHead += n;
  return n;
}

uint32  Stream::flushRwBuffer0(void)
{
  ssize_t  written;
  if (rwBufferHead >= rwBuffer[1])
    {
      written = gff_write(gff, rwBuffer[0], rwBufferSize);
      memcpy(rwBuffer[0], rwBuffer[1], rwBufferHead-rwBuffer[1]);
    }
  else
    {
      written = gff_write(gff, rwBuffer[0], rwBufferHead-rwBuffer[0]);
    }
  rwBufferHead -= written;
  return written;
}

uint32  Stream::flushRwBuffer(void)
{
  ssize_t  written = 0;
  if (rwBufferHead >= rwBuffer[1])
    written += flushRwBuffer0();
  if (rwBufferHead > rwBuffer[0])
    written += flushRwBuffer0();
  return written;
}

///////////////////////////////////////////////////////////////////////////////

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
